#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Tim Doty
#   Simple syntax highlighting editor for PageStream python scripts. Modified/extended
#   version of code bearing the following copyright and license. Distributed as example
#   code for "Rapid GUI Programming with Python and Qt". The python syntax code is
#   lifted nearly verbatim from sandbox.pyw, also by Mark Summerfield, and also with
#   the same copyright and license as indicated below. The qrc_resources file is
#   identical to that provided with the source code for the book.
# Copyright (c) 2007-8 Qtrac Ltd. All rights reserved.
# This program or module is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 2 of the License, or
# version 3 of the License, or (at your option) any later version. It is
# provided for educational purposes and is distributed in the hope that
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
# the GNU General Public License for more details.


# Macros
#   hook into TextEdit event
#   use textChanged() signal?
#   use cursorPositionChange() signal?
#   - basically need to track all keyboard input: cursor keys, delete, backspace,
#     typing
#   - May want to read the article on the undo framework on the publisher website
#     as that had bearing on a macro system
#   - how to bind keys for macros? Or present in toolbar? Or a floating toolbar?

#####
# Check out
# QTextEdit
#   .lineWrapMode()

#####
# Bugs?
#   At some point the handling of non-number bar display was corrupted. Possibly
#   a side effect of the re-write. Further testing and the problem went away. It
#   was characterized by the line numbering not being turned off, the menu checking
#   not updating properly and the blue outline around the text edit window scrolling
#   along with the text (as well as being refreshed around the border). All of this
#   was noted under linux. Hmmm... it is a hit-and-miss bug, it just occured again
#   and went away as quickly.

import os
import sys
import platform
import datetime
from functools import partial
from PyQt4.QtCore import *
from PyQt4.QtGui import *
import qrc_resources

__version__ = "1.1 20110213"

__thisAppName__ = "HC Text Editor"
__companyName__ = "Hammurabi Concepts"
__thisHelpFileName__ = "HCTextEditor.html"

all_windows = []

class NumberBar(QWidget):
    """
        A widget for displaying line numbers for an associated text document.
        Will work for any object supplying the same API as QTextEditor.
        Specifically, the object must provide:
            verticalScrollBar() returning an object providing .maximum() and
                .value(). In the current implementation .minimum() is unused
                but would be necessary if the minimum value were anything
                other than zero.
            document() returning an object providing .blockCount(),
                .findBlockByNumber() and .documentLayout() returning an
                object providing .blockBoundingRect()
    """
    def __init__(self, parent=None):
        super(NumberBar, self).__init__(parent)
        self.editor = None
        # the adjVert was originally necessary, but possibly due to system updates it
        # doesn't appear to have any effect at this time.
        self.adjVert = 5 # default is linux
        self.adjHorz = 5
        self.leftGap = 3
        if platform.system() == 'Darwin':
          self.adjVert = 2
          self.adjHorz = 5
          self.leftGap = 5
        elif platform.system() == 'Windows':
          self.adjVert = 5 # not sure what value should be
          self.adjHorz = 5
        self.visible = False
        self.document = None
        self.documentLayout = None
    def setTextEdit(self, editor):
        self.editor = editor
        self.document = editor.document()
        self.documentLayout = self.document.documentLayout()
    def update(self, *args):
        """
            Updates the number bar to display the current set of numbers.
            Also, adjusts the width of the number bar if necessary.
        """
        if not self.visible:
            self.setFixedWidth(0)
            return QWidget.update(self, *args)
        width = self.fontMetrics().width(str(self.document.blockCount())) + self.adjHorz + self.leftGap
        if self.width() != width:
            self.setFixedWidth(width)
        QWidget.update(self, *args)
    def paintEvent(self, event):
        if not self.visible: return
        coords = []
        fontMetrics = self.fontMetrics()
        # in practice minimum is always zero so assuming that
        maximumY = self.editor.verticalScrollBar().maximum()
        currY = self.editor.verticalScrollBar().value()
        blockCount = self.document.blockCount()
        pageStep = self.editor.verticalScrollBar().pageStep()
        # estimating initial block
        blockNumber = blockCount * currY / (maximumY + pageStep)
        currBlock = self.document.findBlockByNumber(blockNumber)
        #
        blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
        pageBottom = currY + pageStep
        if blockTop >= pageBottom:
            # we guessed high and need to work our way down
            while blockTop > pageBottom:
                currBlock = currBlock.previous()
                blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
            # store coordinates for line number
            lineCount = currBlock.blockNumber() + 1
            x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
            y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
            coords.append( (x, y, str(lineCount)) )
            while blockTop > currY:
                currBlock = currBlock.previous()
                blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
                # store coordinates for line number
                lineCount = currBlock.blockNumber() + 1
                x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
                y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
                coords.append( (x, y, str(lineCount)) )
        elif blockTop <= currY:
            # we guessed low and need to work our way up
            lastBlockTop = blockTop
            while blockTop < currY:
                lastBlockTop = blockTop
                currBlock = currBlock.next()
                blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
            # store coordinates for this and the previous line number (ensures
            # top-most line number is displayed)
            lineCount = currBlock.blockNumber() + 1
            x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
            y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
            if lineCount > 1:
                tmpBlock = currBlock.previous()
                x1 = self.width() - fontMetrics.width(str(lineCount - 1)) - self.adjHorz
                y1 = round(lastBlockTop) - currY + fontMetrics.ascent() + self.adjVert
                if y != y1:
                  coords.append( (x1, y1, str(lineCount - 1)) )
            coords.append( (x, y, str(lineCount)) )
            while blockTop < pageBottom:
                currBlock = currBlock.next()
                if not currBlock.isValid(): break
                blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
                # store coordinates for line number
                lineCount = currBlock.blockNumber() + 1
                x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
                y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
                coords.append( (x, y, str(lineCount)) )
        else:
            # we guessed in the middle
            midBlock = currBlock
            # store coordinates for midBlock
            lineCount = currBlock.blockNumber() + 1
            x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
            y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
            coords.append( (x, y, str(lineCount)) )
            while blockTop > currY and lineCount > 1:
                currBlock = currBlock.previous()
                blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
                # store coordinates for line number
                lineCount = currBlock.blockNumber() + 1
                x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
                y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
                coords.append( (x, y, str(lineCount)) )
            currBlock = midBlock
            while blockTop < pageBottom:
                currBlock = currBlock.next()
                if not currBlock.isValid(): break
                blockTop = self.documentLayout.blockBoundingRect(currBlock).top()
                # store coordinates for line number
                lineCount = currBlock.blockNumber() + 1
                x = self.width() - fontMetrics.width(str(lineCount)) - self.adjHorz
                y = round(blockTop) - currY + fontMetrics.ascent() + self.adjVert
                coords.append( (x, y, str(lineCount)) )
        
        painter = QPainter(self)
        #painter.drawText(0,0,u' ')
        for (x,y,line) in coords:
            painter.drawText(x,y,line)
        painter.end()
        QWidget.paintEvent(self, event)

class HelpContentsDlg(QDialog):
    """
        Displays a help dialog providing information about the application.
    """
    def __init__(self, page, parent=None):
        super(HelpContentsDlg, self).__init__(parent)
        browser = QTextBrowser()
        browser.setSearchPaths([":/"])
        browser.setSource(QUrl(page))
        browser.setOpenExternalLinks(True)
        layout = QVBoxLayout()
        layout.addWidget( browser )
        self.setLayout(layout)

class EditSetDateFormatDlg(QDialog):
    """
        Allows setting the date format string
    """
    def __init__(self, getDateFormatFunc, setDateFormatFunc, parent=None):
        super(EditSetDateFormatDlg, self).__init__(parent)
        self.setDateFormatFunc = setDateFormatFunc
        self.getDateFormatFunc = getDateFormatFunc
        self.dateFormatLineEdit = QLineEdit()
        self.dateFormatLineEdit.setText( getDateFormatFunc() )
        self.dateFormatSpecifierComboBox = QComboBox()
        dateFormatSpecifierList = [
                ["%a: wkdy", "Locale's abbreviated weekday name", "%a"],
                ["%A: Weekday", "Locale's full weekday name", "%A"],
                ["%b: mon", "Locale's abbreviated month name", "%b"],
                ["%B: Month", "Locale's full month name", "%B"],
                ["%c: localized date/time", "Locale's appropriate date and time representation", "%c"],
                ["%d: day", "Day of the month as a decimal number", "%d"],
                [u"%f: μ seconds", "Microsecond as a decimal number, zero-padded on the left", "%f"],
                ["%H: 24H", "Hour (24-hour clock) as a decimal number", "%H"],
                ["%I: 12H", "Hour (12-hour clock) as a decimal number", "%I"],
                ["%j: day of year", "Day of the year as a decimal number", "%j"],
                ["%m: mo", "Month as a decimal number", "%m"],
                ["%M: mi", "Minute as a decimal number", "%M"],
                ["%p: pm", "Locale's equivalent of either AM or PM", "%p"],
                ["%S: seconds", "Second as a decimal number", "%S"],
                ["%U: week number", "Week number of the year as a decimal number, Sunday as first day of week", "%U"],
                ["%w: weekday number", "Weekday as a decimal number", "%w"],
                ["%W: week number", "Week number of the year as a decimal number, Monday as first day of week", "%W"],
                ["%x: localized date", "Locale's appropriate date representation", "%x"],
                ["%X: localized time", "Locale's appropriate time representation", "%X"],
                ["%y: yr", "Last two digits of year", "%y"],
                ["%Y: Year", "Four digit year", "%Y"],
                ["%z: UTC offset", "UTC offset", "%z"],
                ["%Z: tz", "Time zone name", "%Z"],
                ["%%: %", "Literal '%' character", "%%"]
                ]
        self.dateFormatSpecifierToolTipList = []
        for (text, tooltip, format) in dateFormatSpecifierList:
            self.dateFormatSpecifierComboBox.addItem(text, format)
            self.dateFormatSpecifierToolTipList.append(tooltip)
        dateFormatLabel = QLabel('Format String')
        dateFormatLabel.setBuddy(self.dateFormatLineEdit)
        self.dateFormatExampleLabel = QLabel()
        self.dateFormatExampleLabel.setFrameStyle(QFrame.StyledPanel|QFrame.Sunken)
        self.dateFormatExampleLabel.setText( datetime.datetime.today().strftime(self.getDateFormatFunc()) )
        dateFormatExampleLabelLabel = QLabel('Example')
        dateFormatExampleLabelLabel.setBuddy(self.dateFormatExampleLabel)
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Close )
        self.connect(buttonBox, SIGNAL("accepted()"), self.setDateFormat)
        self.connect(buttonBox, SIGNAL("rejected()"), self.close)
        self.connect(self.dateFormatSpecifierComboBox, SIGNAL("currentIndexChanged(int)"), self.insertFormatSpecifier )
        self.connect(self.dateFormatLineEdit, SIGNAL("textChanged(QString)"), self.updateFormatExample )
        self.connect(self.dateFormatSpecifierComboBox, SIGNAL("highlighted(int)"), self.updateToolTip )
        dateFormatLayout = QHBoxLayout()
        dateFormatLayout.addWidget(dateFormatLabel)
        dateFormatLayout.addWidget(self.dateFormatLineEdit)
        dateFormatLayout.addWidget(self.dateFormatSpecifierComboBox)
        exampleLayout = QHBoxLayout()
        exampleLayout.addWidget(dateFormatExampleLabelLabel)
        exampleLayout.addWidget(self.dateFormatExampleLabel)
        buttonBoxLayout = QHBoxLayout()
        buttonBoxLayout.addWidget(buttonBox)
        layout = QVBoxLayout()
        layout.addLayout(dateFormatLayout)
        layout.addLayout(exampleLayout)
        layout.addLayout(buttonBoxLayout)
        self.setLayout(layout)
    def setDateFormat(self):
        self.setDateFormatFunc(str(self.dateFormatLineEdit.text()))
        self.hide()
    def close(self):
        self.dateFormatLineEdit.setText(self.getDateFormatFunc())
        self.hide()
    def insertFormatSpecifier(self, index):
        formatSpecifier = self.dateFormatSpecifierComboBox.itemData(index).toString()
        self.dateFormatLineEdit.insert( formatSpecifier )
    def updateFormatExample(self, dateFormat):
        try:
            #print dateFormat
            dateFormat = str(self.dateFormatLineEdit.text())
            self.dateFormatExampleLabel.setText( datetime.datetime.today().strftime(dateFormat) )
        except ValueError:
            # this happens, for example, when a % is unqualified
            pass
    def updateToolTip(self, index):
        self.dateFormatSpecifierComboBox.setToolTip(self.dateFormatSpecifierToolTipList[index])

class LineEditHistory(QLineEdit):
    """
        Provides a QLineEdit object with an unbounded history of previously
        entered strings that can be selected from by arrowing up or down.
        The history is similar to a stack with each new entered string being
        pushed onto it. A string is considered "entered" whenever the 
        returnPressed() signal is invoked.
        
        The current string is not part of the stack and any editing of the
        displayed string replaces the current string and sets the history
        index to point to the current string instead of the stack.

        The history is automatically updated via the returnPressed() signal,
        but it is the caller's responsibility to call this function at other
        appropriate times, such as when the parent dialog is accepted.
    """
    def __init__(self, parent=None):
        super(LineEditHistory, self).__init__(parent)
        self.historyIndex = 0
        self.historyText = ''
        self.historyList = []
        self.connect(self, SIGNAL("returnPressed()"), self.insertHistory)
        self.connect(self, SIGNAL("textEdited(QString)"), self.updateHistory)
    def insertHistory(self):
        """
            Pushes the current string onto the history stack if it is
            different from the string on the top of the stack.
        """
        if len(self.historyList) == 0 or self.text() != self.historyList[0]:
            self.historyList.insert(0, self.text())
    def event(self, event):
        if event.type() == QEvent.KeyPress:
            historyLength = len(self.historyList)
            if event.key() == Qt.Key_Up:
                self.setHistoryIndex( self.historyIndex + 1 )
                return True
            elif event.key() == Qt.Key_Down:
                self.setHistoryIndex( self.historyIndex - 1 )
                return True
        return QLineEdit.event(self, event)
    def setHistoryText(self, text):
        self.historyText = text
    def clearHistory(self):
        self.historyIndex = 0
        self.historyText = ''
        self.historyList = []
    def setHistoryIndex(self, index):
        """
            Function to safely set the history index to any integer value. It
            is the caller's responsibility to ensure that an integer is passed
            and the function will ensure that the history index is always set
            to a valid value.
            
            Negative values are interpreted as zero, which itself refers to the
            current string. Positive values up to the length of the history
            buffer are allowed. Higher values are set to the length of the
            history buffer. These values indicate the position in the stack
            being referenced.
        """
        lengthHistoryList = len(self.historyList)
        if index < 1 or lengthHistoryList == 0:
            self.historyIndex = 0
            self.setText(self.historyText)
        else:
            if index > lengthHistoryList:
                index = lengthHistoryList
            if self.historyIndex == 0:
                self.historyText = self.text()
            self.historyIndex = index
            self.setText(self.historyList[self.historyIndex - 1])
        self.selectAll()
    def updateHistory(self, text):
        """
            Sets the history index to the current string, called when the
            textEdited() signal is emitted.
        """
        self.historyIndex = 0

class MacroRepeatDlg(QDialog):
    """
        Provides mechanism to select a macro to play an aribtrary number of times.
    """
    def __init__(self, playMacroFunc, parent=None):
        super(MacroRepeatDlg, self).__init__(parent)
        self.playMacroFunc = playMacroFunc
        self.__currMacroName = QString('last recorded')
        self.app = parent
        self.iterSpinBox = QSpinBox()
        self.iterSpinBox.setMinimum(1)
        #self.iterSpinBox.setMaximum(9999)
        iterLabel = QLabel("Number of times")
        iterLabel.setBuddy(self.iterSpinBox)
        self.macroComboBox = QComboBox()
        macroLabel = QLabel("Macro to play")
        macroLabel.setBuddy(self.macroComboBox)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Close)
        self.buttonBox.addButton("Repeat", QDialogButtonBox.AcceptRole)
        self.connect(self.buttonBox, SIGNAL("accepted()"), self.playMacroFunc)
        self.connect(self.buttonBox, SIGNAL("rejected()"), self.close)
        iterRow = QHBoxLayout()
        iterRow.addWidget(iterLabel)
        iterRow.addWidget(self.iterSpinBox)
        macroRow = QHBoxLayout()
        macroRow.addWidget(macroLabel)
        macroRow.addWidget(self.macroComboBox)
        buttonRow = QHBoxLayout()
        buttonRow.addWidget(self.buttonBox)
        layout = QVBoxLayout()
        layout.addLayout( macroRow )
        layout.addLayout( iterRow )
        layout.addWidget( self.buttonBox )
        self.update()
        self.setLayout(layout)
    def update(self):
        self.macroComboBox.clear()
        if len(self.app.editor.macroKeyPresses) > 0:
            self.macroComboBox.addItem(self.__currMacroName)
        for macro in self.app.storedMacroList:
            self.macroComboBox.addItem(macro)

# not currently used, but then again keyboard shortcuts to trigger a macro aren't either...
class MacroObject():
    """
        Container for a macro including a display name, the sequence of key press events that comprise it,
        and any keyboard shortcut which is to invoke the macro.
    """
    def __init__(self, name, eventSequence=None, shortcut=None):
        self.name = name
        self.events = []
        self.shortcut = shortcut
        for event in eventSequence:
            self.events.append( QKeyPress(event) )
    def addEvent(self, event):
        self.events.append( QKeyPress(event) )
    def clear(self):
        self.name = QString('')
        self.events.clear()
        self.shortcut = None
    def setEvents(self, eventSequence):
        self.events.clear()
        for event in eventSequence:
            self.events.append( QKeyPress(event) )
    def setName(self, name):
        self.name = name
    def setShortcut(self, shortcut):
        self.shortcut = shortcut
    def pop(self):
        return self.events.pop()

class MacroManageDlg(QDialog):
    """
        Interface to name (and store) current macro, rename or remove
        existing stored macros. Allows inspection of macro results, inasmuch
        as that even makes sense. Can specify a text to use as basis for the
        visualization.
    """
    def __init__(self, parent=None):
        super(MacroManageDlg, self).__init__(parent)
        self.app = parent
        self.macroBackStore = {}
        self.selectedMacro = QString('')
        self.__currMacroName = QString("current macro")
        self.macroListWidget = QListWidget()
        macroLabel = QLabel("Existing Macros")
        macroLabel.setBuddy(self.macroListWidget)
        macroCol = QVBoxLayout()
        macroCol.addWidget(macroLabel)
        macroCol.addWidget(self.macroListWidget)
        self.nameLineEdit = QLineEdit()
        nameLabel = QLabel("Macro Name")
        nameLabel.setBuddy(self.nameLineEdit)
        self.macroTextEdit = QTextEdit()
        self.macroTextEdit.setReadOnly(True)
        detailCol = QVBoxLayout()
        detailCol.addWidget(nameLabel)
        detailCol.addWidget(self.nameLineEdit)
        detailCol.addWidget(self.macroTextEdit)
        macroLayout = QHBoxLayout()
        macroLayout.addLayout(macroCol)
        macroLayout.addLayout(detailCol)
        # FIXME: add button to make currently selected stored macro the current editor macro
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Reset | QDialogButtonBox.Close )
        deleteButton = buttonBox.addButton("Delete", QDialogButtonBox.DestructiveRole)
        currentButton = buttonBox.addButton("Make Current", QDialogButtonBox.ActionRole)
        layout = QVBoxLayout()
        layout.addLayout(macroLayout)
        layout.addWidget(buttonBox)
        self.setLayout(layout)
        self.connect(buttonBox, SIGNAL("accepted()"), self.save)
        self.connect(buttonBox.button(QDialogButtonBox.Reset), SIGNAL("clicked()"), self.reset)
        self.connect(deleteButton, SIGNAL("clicked()"), self.deleteMacro)
        self.connect(currentButton, SIGNAL("clicked()"), self.makeCurrent)
        self.connect(buttonBox, SIGNAL("rejected()"), self.close)
        self.connect(self.macroListWidget, SIGNAL("itemSelectionChanged()"), self.updateMacroWidgets)
        self.connect(self.nameLineEdit, SIGNAL("editingFinished()"), self.storeName)
        self.update()
    def deleteMacro(self):
        """
            Used to delete any stored macro, but has no effect when used on the editor's
            current recorded macro
        """
        if self.selectedMacro == self.__currMacroName: return
        self.app.storedMacroList.pop(self.selectedMacro)
        self.selectedMacro = None
        self.macroListWidget.setCurrentRow(0)
        self.update()
    def makeCurrent(self):
        self.app.editor.macroKeyPresses = []
        for event in self.app.storedMacroList[self.selectedMacro]:
            self.app.editor.macroKeyPresses.append(QKeyEvent(event))
        self.update()
    def backStore(self):
        """
            Stores the current macros so that they can be restored later via reset
        """
        self.macroBackStore.clear()
        for macro in self.app.storedMacroList:
            self.macroBackStore[macro] = []
            for event in self.app.storedMacroList[macro]:
                self.macroBackStore[macro].append(QKeyEvent(event))
    def reset(self):
        """
            Restores any back stored macros, replacing the currently defined macro set
        """
        self.selectedMacro = None
        self.app.storedMacroList.clear()
        for macro in self.macroBackStore:
            self.app.storedMacroList[macro] = []
            for event in self.macroBackStore[macro]:
                self.app.storedMacroList[macro].append(QKeyEvent(event))
        self.update()
    def update(self):
        self.macroListWidget.clear()
        if len(self.app.editor.macroKeyPresses) > 0:
            self.macroListWidget.addItem(self.__currMacroName)
        for macro in self.app.storedMacroList:
            self.macroListWidget.addItem(macro)
        self.macroTextEdit.clear()
    def cancel(self):
        self.reset()
        self.hide()
    def close(self):
        self.reset()
        self.hide()
    def save(self):
        self.storeName()
        self.macroBackStore.clear()
        self.hide()
    def updateMacroWidgets(self):
        item = self.macroListWidget.currentItem()
        if item: self.selectedMacro = item.text()
        self.nameLineEdit.setText(self.selectedMacro)
        if self.selectedMacro == self.__currMacroName:
            self.setEventText(self.app.editor.macroKeyPresses)
        else:
            self.setEventText(self.app.storedMacroList[self.selectedMacro])
    def storeName(self):
        newMacroName = self.nameLineEdit.text()
        if newMacroName == self.__currMacroName: return
        if self.selectedMacro == self.__currMacroName:
            self.app.storedMacroList[newMacroName] = []
            for event in self.app.editor.macroKeyPresses:
                self.app.storedMacroList[newMacroName].append(QKeyEvent(event))
        elif newMacroName != self.selectedMacro:
            self.app.storedMacroList[newMacroName] = self.app.storedMacroList.pop(self.selectedMacro)
            self.selectedMacro = newMacroName
        self.macroListWidget.setCurrentItem(QListWidgetItem(newMacroName))
        self.update()
    def setEventText(self, eventList):
        self.macroTextEdit.clear()
        for event in eventList:
            text = str(event.text())
            if text > "\x20" and text < "\x7f":
                self.macroTextEdit.insertHtml("<tt>" + event.text() + "</tt>")
            elif event.key() == Qt.Key_Space:
                self.macroTextEdit.insertHtml("<b>[Space]</b>")
            elif event.key() == Qt.Key_Tab:
                self.macroTextEdit.insertHtml("<b>[Tab]</b>")
            elif event.key() == Qt.Key_Backspace:
                self.macroTextEdit.insertHtml("<b>[Backspace]</b>")
            elif event.key() == Qt.Key_Delete:
                self.macroTextEdit.insertHtml("<b>[Delete]</b>")
            elif event.key() == Qt.Key_Return:
                self.macroTextEdit.insertHtml("<b>[Return]</b>")
            elif event.key() == Qt.Key_Enter:
                self.macroTextEdit.insertHtml("<b>[Enter]</b>")
            elif event.key() == Qt.AltModifier:
                self.macroTextEdit.insertHtml("<b>[Alt]</b>")
            elif event.key() == Qt.MetaModifier:
                self.macroTextEdit.insertHtml("<b>[Meta]</b>")
            elif event.key() == Qt.Key_Escape:
                self.macroTextEdit.insertHtml("<b>[Escape]</b>")
            elif event.key() == Qt.Key_CapsLock:
                self.macroTextEdit.insertHtml("<b>[CapsLock]</b>")
            elif event.key() == Qt.ShiftModifier:
                self.macroTextEdit.insertHtml("<b>[Shift]</b>")
            elif event.key() == Qt.ControlModifier:
                self.macroTextEdit.insertHtml("<b>[Control]</b>")
            elif event.key() == Qt.Key_Meta:
                if platform.system() == 'Darwin':
                    self.macroTextEdit.insertHtml("<b>[Control]</b>")
                else:
                    self.macroTextEdit.insertHtml("<b>[Windows]</b>")
            elif event.key() == Qt.Key_Shift:
                self.macroTextEdit.insertHtml("<b>[Shift]</b>")
            elif event.key() == Qt.Key_Control:
                if platform.system() == 'Darwin':
                    self.macroTextEdit.insertHtml("<b>[Command]</b>")
                else:
                    self.macroTextEdit.insertHtml("<b>[Control]</b>")
            elif event.key() == Qt.Key_Alt:
                if platform.system() == 'Darwin':
                    self.macroTextEdit.insertHtml("<b>[Option]</b>")
                else:
                    self.macroTextEdit.insertHtml("<b>[Alt]</b>")
            elif event.key() == Qt.Key_AltGr:
                self.macroTextEdit.insertHtml("<b>[AltGr]</b>")
            elif event.key() == Qt.Key_End:
                self.macroTextEdit.insertHtml("<b>[End]</b>")
            elif event.key() == Qt.Key_Backtab:
                self.macroTextEdit.insertHtml("<b>[Backtab]</b>")
            elif event.key() == Qt.Key_Insert:
                self.macroTextEdit.insertHtml("<b>[Insert]</b>")
            elif event.key() == Qt.Key_Pause:
                self.macroTextEdit.insertHtml("<b>[Pause]</b>")
            elif event.key() == Qt.Key_Print:
                self.macroTextEdit.insertHtml("<b>[Print]</b>")
            elif event.key() == Qt.Key_SysReq:
                self.macroTextEdit.insertHtml("<b>[SysReq]</b>")
            elif event.key() == Qt.Key_Clear:
                self.macroTextEdit.insertHtml("<b>[Clear]</b>")
            elif event.key() == Qt.Key_Home:
                self.macroTextEdit.insertHtml("<b>[Home]</b>")
            elif event.key() == Qt.Key_PageUp:
                self.macroTextEdit.insertHtml("<b>[PageUp]</b>")
            elif event.key() == Qt.Key_PageDown:
                self.macroTextEdit.insertHtml("<b>[PageDown]</b>")
            elif event.key() == Qt.Key_NumLock:
                self.macroTextEdit.insertHtml("<b>[NumLock]</b>")
            elif event.key() == Qt.Key_ScrollLock:
                self.macroTextEdit.insertHtml("<b>[ScrollLock]</b>")
            elif event.key() == Qt.Key_Help:
                self.macroTextEdit.insertHtml("<b>[Help]</b>")
            else:
                self.macroTextEdit.insertHtml("<b>[#" + str(event.key()) + "]</b>")

class SearchFindDlg(QDialog):
    def __init__(self, findFunc, parent=None):
        super(SearchFindDlg, self).__init__(parent)
        self.findFunc = findFunc
        self.findField = LineEditHistory()
        findFieldLabel = QLabel("&Find:")
        findFieldLabel.setBuddy(self.findField)
        self.caseSensitive = QCheckBox("Case &sensitive")
        self.findBackward = QCheckBox("&Backwards")
        self.matchComboBox = QComboBox()
        self.matchComboBox.addItem("Substring")
        self.matchComboBox.addItem("Whole words only")
        self.matchComboBox.addItem("Regular expression")
        findMatchLabel = QLabel("&Match by")
        findMatchLabel.setBuddy(self.matchComboBox)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Close)
        self.buttonBox.addButton("Find", QDialogButtonBox.AcceptRole)
        self.connect(self.buttonBox, SIGNAL("accepted()"), self.find)
        self.connect(self.buttonBox, SIGNAL("rejected()"), self.close)
        findRow = QHBoxLayout()
        findRow.addWidget(findFieldLabel)
        findRow.addWidget(self.findField)
        optRow = QHBoxLayout()
        optRow.addWidget(self.caseSensitive)
        optRow.addWidget(self.findBackward)
        optRow.addWidget(self.matchComboBox)
        layout = QVBoxLayout()
        layout.addLayout( findRow )
        layout.addLayout( optRow )
        layout.addWidget( self.buttonBox )
        self.setLayout(layout)
    def close(self):
        self.hide()
    def resetText(self):
        self.findField.setHistoryText('')
        self.findField.setHistoryIndex(1)
    def find(self):
        self.findField.insertHistory()
        self.findFunc()

class SearchReplaceDlg(QDialog):
    def __init__(self, replaceFunc, parent=None):
        super(SearchReplaceDlg, self).__init__(parent)
        self.app = parent
        self.replaceFunc = replaceFunc
        self.findField = LineEditHistory()
        findFieldLabel = QLabel("&Find:")
        findFieldLabel.setBuddy(self.findField)
        self.replaceField = LineEditHistory()
        replaceFieldLabel = QLabel("&Replace:")
        replaceFieldLabel.setBuddy(self.replaceField)
        self.caseSensitive = QCheckBox("Case &sensitive")
        self.findBackward = QCheckBox("&Backwards")
        self.matchComboBox = QComboBox()
        self.matchComboBox.addItem("Substring")
        self.matchComboBox.addItem("Whole words only")
        self.matchComboBox.addItem("Regular expression")
        findMatchLabel = QLabel("&Match by")
        findMatchLabel.setBuddy(self.matchComboBox)
        buttonBox = QDialogButtonBox(QDialogButtonBox.Close)
        buttonReplace = buttonBox.addButton("Replace", QDialogButtonBox.AcceptRole)
        buttonReplaceAll = buttonBox.addButton("Replace All", QDialogButtonBox.AcceptRole)
        self.connect(buttonBox, SIGNAL("rejected()"), self.close)
        self.connect(buttonReplace, SIGNAL("clicked()"), partial(self.replace, False))
        self.connect(buttonReplaceAll, SIGNAL("clicked()"), partial(self.replace, True))
        findRow = QHBoxLayout()
        findRow.addWidget(findFieldLabel)
        findRow.addWidget(self.findField)
        replaceRow = QHBoxLayout()
        replaceRow.addWidget(replaceFieldLabel)
        replaceRow.addWidget(self.replaceField)
        optRow = QHBoxLayout()
        optRow.addWidget(self.caseSensitive)
        optRow.addWidget(self.findBackward)
        optRow.addWidget(self.matchComboBox)
        layout = QVBoxLayout()
        layout.addLayout( findRow )
        layout.addLayout( replaceRow )
        layout.addLayout( optRow )
        layout.addWidget( buttonBox )
        self.setLayout(layout)
    def close(self):
        self.hide()
    def resetText(self):
        self.findField.setHistoryText('')
        self.findField.setHistoryIndex(1)
        self.replaceField.setHistoryText('')
        self.replaceField.setHistoryIndex(1)
    def replace(self, replaceAll=False):
        self.findField.insertHistory()
        self.replaceField.insertHistory()
        self.replaceFunc(replaceAll=replaceAll)

class ToolsPrefsDlg(QDialog):
    def __init__(self, updateProfileFunc, restoreDefaultProfileFunc, parent=None):
        super(ToolsPrefsDlg, self).__init__(parent)
        self.app = parent
        self.colorSwatchWidth = 16
        self.colorSwatchHeight = 16
        self.highlightLineColor = QColor()
        self.profileComboBox = QComboBox()
        for profileName in self.app.profiles:
            self.profileComboBox.addItem(self.app.profiles[profileName].name, profileName)
        profileLabel = QLabel("&Profile to edit")
        profileLabel.setBuddy(self.profileComboBox)
        profileRow = QHBoxLayout()
        profileRow.addWidget(profileLabel)
        profileRow.addWidget(self.profileComboBox)
        # 1. not part of profiles at this time
        # 2. needs to have a spinbox for point size
        self.fontComboBox = QFontComboBox()
        fontLabel = QLabel("&Font")
        fontLabel.setBuddy(self.fontComboBox)
        self.fontSizeSpinBox = QSpinBox()
        self.fontSizeSpinBox.setMinimum(6)
        self.fontSizeSpinBox.setMaximum(255)
        fontRow = QHBoxLayout()
        fontRow.addWidget(fontLabel)
        fontRow.addWidget(self.fontComboBox)
        fontRow.addWidget(self.fontSizeSpinBox)
        self.tabstopSpinBox = QSpinBox()
        self.tabstopSpinBox.setMinimum(1)
        self.tabstopSpinBox.setMaximum(99)
        tabstopLabel = QLabel("&Tabstop")
        tabstopLabel.setBuddy(self.tabstopSpinBox)
        self.tabToSpaceCheckBox = QCheckBox("&Convert tabs to spaces")
        self.showLineNumbersCheckBox = QCheckBox("&Show line numbers")
        self.maintainIndentComboBox = QComboBox()
        for method in ['None', 'Normal', 'Smart']:
            self.maintainIndentComboBox.addItem(method)
        maintainIndentLabel = QLabel("&Automatic Indent")
        maintainIndentLabel.setBuddy(self.maintainIndentComboBox)
        maintainIndentRow = QHBoxLayout()
        maintainIndentRow.addWidget(maintainIndentLabel)
        maintainIndentRow.addWidget(self.maintainIndentComboBox)
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Close | QDialogButtonBox.RestoreDefaults)
        tabstopRow = QHBoxLayout()
        tabstopRow.addWidget(tabstopLabel)
        tabstopRow.addWidget(self.tabstopSpinBox)
        self.wordWrapComboBox = QComboBox()
        for method in self.app.wordWrapMethods:
            self.wordWrapComboBox.addItem(method, self.app.wordWrapMethods[method]['data'])
        wordWrapLabel = QLabel("&Word wrap")
        wordWrapLabel.setBuddy(self.wordWrapComboBox)
        wordWrapRow = QHBoxLayout()
        wordWrapRow.addWidget(wordWrapLabel)
        wordWrapRow.addWidget(self.wordWrapComboBox)
        self.styleListWidget = QListWidget()
        styleLabel = QLabel("Syntax &Highlight Formats")
        styleLabel.setBuddy(self.styleListWidget)
        self.styleColorPushButton = QPushButton()
        self.styleBoldCheckBox = QCheckBox('Bold')
        self.styleItalicCheckBox = QCheckBox('Italic')
        styleInputRow = QHBoxLayout()
        styleInputRow.addWidget(self.styleColorPushButton)
        styleInputRow.addWidget(self.styleBoldCheckBox)
        styleInputRow.addWidget(self.styleItalicCheckBox)
        self.highlightCheckBox = QCheckBox('Highlight Current Line')
        self.highlightColorPushButton = QPushButton()
        highlightRow = QHBoxLayout()
        highlightRow.addWidget(self.highlightCheckBox)
        highlightRow.addWidget(self.highlightColorPushButton)
        styleLayout = QVBoxLayout()
        styleLayout.addWidget(styleLabel)
        styleLayout.addWidget(self.styleListWidget)
        styleLayout.addLayout(styleInputRow)
        styleLayout.addLayout(highlightRow)
        layout = QVBoxLayout()
        layout.addLayout(profileRow)
        layout.addLayout(fontRow)
        layout.addLayout(tabstopRow)
        layout.addWidget(self.tabToSpaceCheckBox)
        layout.addWidget(self.showLineNumbersCheckBox)
        layout.addLayout(maintainIndentRow)
        layout.addLayout(wordWrapRow)
        layout.addWidget(buttonBox)
        dlgLayout = QHBoxLayout()
        dlgLayout.addLayout(layout)
        dlgLayout.addLayout(styleLayout)
        self.setLayout(dlgLayout)
        self.connect(self.profileComboBox, SIGNAL("currentIndexChanged(int)"), self.indexChanged)
        self.connect(buttonBox.button(QDialogButtonBox.Ok), SIGNAL("clicked()"), updateProfileFunc)
        self.connect(buttonBox.button(QDialogButtonBox.RestoreDefaults), SIGNAL("clicked()"), restoreDefaultProfileFunc)
        self.connect(buttonBox, SIGNAL("rejected()"), self.close)
        self.connect(self.styleColorPushButton, SIGNAL("clicked()"), self.styleColorDialog)
        self.connect(self.styleListWidget, SIGNAL("itemSelectionChanged()"), self.updateStyleWidgets)
        self.connect(self.styleBoldCheckBox, SIGNAL("stateChanged(int)"), self.updateStyleData)
        self.connect(self.styleItalicCheckBox, SIGNAL("stateChanged(int)"), self.updateStyleData)
        self.connect(self.highlightColorPushButton, SIGNAL("clicked()"), self.highlightColorDialog)
        self.indexChanged(0)
    def close(self):
        self.hide()
    def indexChanged(self, index):
        currProfile = unicode(self.profileComboBox.itemData(index).toString())
        self.fontComboBox.setCurrentFont(self.app.profiles[currProfile].font)
        self.fontSizeSpinBox.setValue(self.app.profiles[currProfile].font.pointSize())
        self.tabstopSpinBox.setValue(self.app.profiles[currProfile].tabStop)
        self.tabToSpaceCheckBox.setChecked(self.app.profiles[currProfile].convertTabs)
        self.showLineNumbersCheckBox.setChecked(self.app.profiles[currProfile].showLineNumbers)
        self.maintainIndentComboBox.setCurrentIndex(
            self.maintainIndentComboBox.findText(self.app.profiles[currProfile].maintainIndent))
        self.wordWrapComboBox.setCurrentIndex(
            self.wordWrapComboBox.findText(self.app.profiles[currProfile].wordWrap))
        self.highlightLineColor = self.app.profiles[currProfile].highlightLineColor
        currColorPixmap = QPixmap(self.colorSwatchWidth, self.colorSwatchHeight)
        currColorPixmap.fill(self.highlightLineColor)
        currColorIcon = QIcon(currColorPixmap)
        self.highlightColorPushButton.setIcon(currColorIcon)
        self.highlightCheckBox.setChecked(self.app.profiles[currProfile].highlightCurrentLine)
        # remove existing entries from self.styleListWidget
        self.styleListWidget.clear()
        formats = self.app.profiles[currProfile].formats
        for formatName in formats:
            currColor = formats[formatName]['color']
            currColorPixmap = QPixmap(self.colorSwatchWidth, self.colorSwatchHeight)
            currColorPixmap.fill(currColor)
            currColorIcon = QIcon(currColorPixmap)
            currItem = QListWidgetItem(currColorIcon, formatName)
            currFont = currItem.font()
            currFont.setBold(formats[formatName]['bold'])
            currFont.setItalic(formats[formatName]['italic'])
            if self.app.fontOveride:
                currItem.setFont(self.app.fontOveride)
            else:
                currItem.setFont(currFont)
            currItem.setData(32, currColor)
            currItem.setData(33, formats[formatName]['bold'])
            currItem.setData(34, formats[formatName]['italic'])
            self.styleListWidget.addItem(currItem)
        self.styleListWidget.setCurrentRow(0)
    def styleColorDialog(self):
        currItem = self.styleListWidget.currentItem()
        color = QColor(currItem.data(32))
        color = QColorDialog().getColor(color)
        currItem.setData(32, color)
        currColorPixmap = QPixmap(self.colorSwatchWidth, self.colorSwatchHeight)
        currColorPixmap.fill(color)
        currColorIcon = QIcon(currColorPixmap)
        currItem.setIcon(currColorIcon)
        self.styleColorPushButton.setIcon(currColorIcon)
    def highlightColorDialog(self):
        self.highlightLineColor = QColorDialog().getColor(self.highlightLineColor)
        currColorPixmap = QPixmap(self.colorSwatchWidth, self.colorSwatchHeight)
        currColorPixmap.fill(self.highlightLineColor)
        currColorIcon = QIcon(currColorPixmap)
        self.highlightColorPushButton.setIcon(currColorIcon)
    def updateStyleWidgets(self):
        currItem = self.styleListWidget.currentItem()
        if currItem:
            color = QColor(currItem.data(32))
            bold = currItem.data(33).toBool()
            italic = currItem.data(34).toBool()
            colorPixmap = QPixmap(self.colorSwatchWidth, self.colorSwatchHeight)
            colorPixmap.fill(color)
            colorIcon = QIcon(colorPixmap)
            self.styleColorPushButton.setIcon(colorIcon)
            self.styleBoldCheckBox.setChecked(bold)
            self.styleItalicCheckBox.setChecked(italic)
    def updateStyleData(self):
        currItem = self.styleListWidget.currentItem()
        if currItem:
            bold = self.styleBoldCheckBox.isChecked()
            italic = self.styleItalicCheckBox.isChecked()
            currItem.setData(33, bold)
            currItem.setData(34, italic)
            currFont = currItem.font()
            currFont.setBold(bold)
            currFont.setItalic(italic)
            if self.app.fontOveride:
                currItem.setFont(self.app.fontOveride)
            else:
                currItem.setFont(currFont)

class ToolsTabDlg(QDialog):
    def __init__(self, storeTabStopFunc, parent=None):
        super(ToolsTabDlg, self).__init__(parent)
        self.tabstopSpinBox = QSpinBox()
        self.tabstopSpinBox.setMinimum(1)
        self.tabstopSpinBox.setMaximum(99)
        tabstopLabel = QLabel("&Tabstop")
        tabstopLabel.setBuddy(self.tabstopSpinBox)
        self.tabToSpaceCheckBox = QCheckBox("&Convert tabs to spaces")
        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Close)
        tabstopRow = QHBoxLayout()
        tabstopRow.addWidget(tabstopLabel)
        tabstopRow.addWidget(self.tabstopSpinBox)
        layout = QVBoxLayout()
        layout.addLayout(tabstopRow)
        layout.addWidget(self.tabToSpaceCheckBox)
        layout.addWidget(buttonBox)
        self.setLayout(layout)
        self.connect(buttonBox.button(QDialogButtonBox.Ok), SIGNAL("clicked()"), storeTabStopFunc)
        self.connect(buttonBox, SIGNAL("rejected()"), self.close)
    def close(self):
        self.hide()
    def setTabstop(self, tabstop=1):
        self.tabstopSpinBox.setValue( tabstop )
    def setConvertTabs(self, convert=False):
        self.tabToSpaceCheckBox.setChecked( convert )

class PythonHighlighter(QSyntaxHighlighter):

    Rules = []
    Formats = {}

    def __init__(self, parent=None):
        super(PythonHighlighter, self).__init__(parent)
        self.useBlocks = False
        baseFormat = QTextCharFormat()
        (name, color, bold, italic) = ("normal", "#000000", False, False)
        format = QTextCharFormat(baseFormat)
        format.setForeground(QColor(color))
        if bold:
            format.setFontWeight(QFont.Bold)
        format.setFontItalic(italic)
        self.Formats[name] = format
        # for Python syntax highlighting
        self.stringRe = QRegExp(r"""(:?"["]".*"["]"|'''.*''')""")
        self.stringRe.setMinimal(True)
        self.tripleSingleRe = QRegExp(r"""'''(?!")""")
        self.tripleDoubleRe = QRegExp(r'''"""(?!')''')
    def highlightBlock(self, text):
        # need to check if text being highlighted is current block
        # if so, and highlightCurrentLine is true then use alternate
        # format for setFormat() where .setBackground() is
        # highlightLineColor
        if self.useBlocks:
            NORMAL, TRIPLESINGLE, TRIPLEDOUBLE, ERROR = range(4)
            textLength = text.length()
            prevState = self.previousBlockState()
            self.setFormat(0, textLength, self.Formats["normal"])
            if text.startsWith("Traceback") or text.startsWith("Error: "):
                self.setCurrentBlockState(ERROR)
                self.setFormat(0, textLength, self.Formats["error"])
                return
            if (prevState == ERROR and
                not (text.startsWith(sys.ps1) or text.startsWith("#"))):
                self.setCurrentBlockState(ERROR)
                self.setFormat(0, textLength, self.Formats["error"])
                return
            for regex, format in self.Rules:
                i = regex.indexIn(text)
                while i >= 0:
                    length = regex.matchedLength()
                    self.setFormat(i, length, self.Formats[format])
                    i = regex.indexIn(text, i + length)
            self.setCurrentBlockState(NORMAL)
            if self.stringRe.indexIn(text) != -1:
                return
            # This is fooled by triple quotes inside single quoted strings
            for i, state in ((self.tripleSingleRe.indexIn(text),
                              TRIPLESINGLE),
                             (self.tripleDoubleRe.indexIn(text),
                              TRIPLEDOUBLE)):
                if self.previousBlockState() == state:
                    if i == -1:
                        i = text.length()
                        self.setCurrentBlockState(state)
                    self.setFormat(0, i + 3, self.Formats["string"])
                elif i > -1:
                    self.setCurrentBlockState(state)
                    self.setFormat(i, text.length(), self.Formats["string"])
        else:
            textLength = text.length()
            self.setFormat(0, textLength, self.Formats["normal"])
            for regex, format in self.Rules:
                i = regex.indexIn(text)
                while i >= 0:
                    length = regex.matchedLength()
                    self.setFormat(i, length, self.Formats[format])
                    i = regex.indexIn(text, i + length)
    def rehighlight(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        QSyntaxHighlighter.rehighlight(self)
        QApplication.restoreOverrideCursor()
    def setRules(self, rules):
        self.Rules = []
        for rule in rules:
            self.Rules.append(rule)
    def setFormats(self, formats):
        self.Formats.clear()
        baseFormat = QTextCharFormat()
        for formatName in formats.keys():
            format = QTextCharFormat(baseFormat)
            format.setForeground(formats[formatName]['color'])
            if formats[formatName]['bold']:
                format.setFontWeight(QFont.Bold)
            format.setFontItalic(formats[formatName]['italic'])
            self.Formats[formatName] = format


class TextEdit(QTextEdit):
    def __init__(self, parent=None):
        super(TextEdit, self).__init__(parent)
        self.tabstop = 10
        self.convertTabs = False
        self.tabCount = 0
        self.tabstopWidthFactor = 8
        self.setAcceptRichText(False)
        self.app = None
        self.lastBlock = None
        self.lastCursor = None
        self.highlightCurrentLine = False
        self.highlightLineColor = QColor('#FFFFAA')
        self.searchText = []
        self.replaceText = []
        self.macroKeyPresses = []
        self.macroRecord = False
        self.maintainIndent = None
    def getIndent(self, text):
        tabCount = 0
        spaceCount = 0
        for char in text:
            if char == ' ': spaceCount += 1
            elif char == '\n': tabCount += 1
            else: break
        lastChar = ''
        if text != '':
            lastChar = text[-1]
        return tabCount, spaceCount, lastChar
    def event(self, event):
        if event.type() == QEvent.KeyPress:
            if self.macroRecord:
                self.macroKeyPresses.append(QKeyEvent(event))
            if event.key() == Qt.Key_Tab:
                if self.convertTabs:
                    cursor = self.textCursor()
                    cursor.insertText(" " * self.tabstop)
                    self.updateStatusMessage()
                    return True
            elif event.key() == Qt.Key_Backspace:
                if self.convertTabs:
                    cursor = self.textCursor()
                    if not cursor.hasSelection():
                        index = cursor.position() - cursor.block().position()
                        (tabCount, spaceCount, lastChar) = self.getIndent(cursor.block().text())
                        spaceCount += self.tabstop * tabCount
                        if spaceCount > 0 and index <= spaceCount and index > 0:
                            for i in range(spaceCount % self.tabstop, self.tabstop):
                                cursor.deletePreviousChar()
                            self.updateStatusMessage()
                            return True
            elif event.key() == Qt.Key_Return:
                # consider using returnPressed() signal for this
                if self.maintainIndent == 'Normal':
                    cursor = self.textCursor()
                    (tabCount, spaceCount, lastChar) = self.getIndent(cursor.block().text())
                    cursor.insertBlock()
                    if self.convertTabs:
                        spaceCount += self.tabstop * tabCount
                        cursor.insertText(" " * spaceCount)
                    else:
                        cursor.insertText("\t" * tabCount)
                    self.updateStatusMessage()
                    return True
                elif self.maintainIndent == 'Smart':
                    # note that there is no danger for negative spacing when reducing the tabstop
                    # because when multiplying a character by a negative number the result is an
                    # empty string
                    cursor = self.textCursor()
                    text = cursor.block().text()
                    (tabCount, spaceCount, lastChar) = self.getIndent(text)
                    profileName = unicode(self.app.profileGroup.checkedAction().data().toString())
                    if profileName == 'python':
                        if lastChar == ':' and cursor.atBlockEnd(): tabCount += 1
                        elif text == ' ' * spaceCount or text == '\t' * tabCount: tabCount -= 1
                    elif profileName == 'html':
                        if (text.endsWith("<ul>") or \
                            text.endsWith("<dl>") or \
                            text.endsWith("<table>") or \
                            text.endsWith("<tr>") or \
                            text.endsWith("<head>") or \
                            text.endsWith("<body>") or \
                            text.endsWith("<p>")) and cursor.atBlockEnd(): tabCount += 1
                        # it would work better to delete by one tabstop in current block
                        # then use that reduced indent for the next block. however, if do it that
                        # way can lead to undesirable outdents by repeatedly hitting return after
                        # either accept that or build in greater intelligence, but that would have
                        # to backtrack and correctly identify the opening tag's proper indent. seems
                        # like a lot of room for trouble and perhaps simply better off allowing the
                        # one bad case to exist
                        elif (text.endsWith("</ul>") or \
                            text.endsWith("</dl>") or \
                            text.endsWith("</table>") or \
                            text.endsWith("</tr>") or \
                            text.endsWith("</head>") or \
                            text.endsWith("</body>") or \
                            text.endsWith("</p>")) and cursor.atBlockEnd(): tabCount -= 1
                    elif profileName == 'c_like':
                        if lastChar == '{' and cursor.atBlockEnd(): tabCount += 1
                        # alternatively, for the following delete by one tabstop in current block
                        # then use that reduced indent for the next block
                        elif lastChar == '}' and cursor.atBlockEnd(): tabCount -= 1
                    else:
                        # no adjustments: the same as for 'Normal' maintainIndent
                        pass
                    cursor.insertBlock()
                    if self.convertTabs:
                        spaceCount += self.tabstop * tabCount
                        cursor.insertText(" " * spaceCount)
                    else:
                        cursor.insertText("\t" * tabCount)
                    self.updateStatusMessage()
                    return True
            else:
                pass
        self.updateStatusMessage()
        return QTextEdit.event(self, event)
        
    def updateStatusMessage(self):
        status = self.app.statusBar()
        cursor = self.textCursor()
        position = cursor.position() + 1
        lineNumber = cursor.blockNumber() + 1
        column = position - cursor.block().position()
        lineCount = self.document().lineCount()
        characterCount = self.document().characterCount()
        message = "character %s/%s  line %s/%s  column %s" % (position, characterCount, lineNumber, lineCount, column)
        status.showMessage(message)
        
    def setTabStop(self, tabstop):
        self.tabstop = int(tabstop) if int(tabstop) > 1 else 8
        self.setTabStopWidth(self.tabstop * self.tabstopWidthFactor)
    def getTabStop(self):
        return self.tabstop
    def setConvertTabs(self, convertTabs):
        self.convertTabs = bool(convertTabs)
    def getConvertTabs(self):
        return self.convertTabs
    def lineNumberAreaWidth(self):
        maxNum = qMax(q, self.blockCount())
        space = 3 + fontMetrics().width(QLatin1Char('9')) * len(str(maxNum))
        return space

class prefsProfile():
    """Used to store text editor profiles, specifically aimed at particular file
       formats. Allows specification of full application preferences to provide 
       complete customization."""
    def __init__(self, name, filetype=None, ext=QRegExp("^$")):
        self.name = name
        self.filetype = filetype if filetype is not None else "All files (*.*)"
        self.ext = ext
        self.font = None
        self.fixedPitch = False
        self.tabStop = 10
        self.convertTabs = False
        self.maintainIndent = 'None' # 'Normal' or 'Smart'
        self.useBlocks = False
        self.formats = {}
        self.rules = []
        self.showLineNumbers = False
        self.wordWrap = "Word Wrap"
        self.action = None
        self.highlightCurrentLine = False
        self.highlightLineColor = QColor('#FFFFAA')
    def setFormatAttr(self, name, color='#000000', bold=False, italic=False):
        self.formats[name] = { 'color': QColor(color), 'bold': bold, 'italic': italic }
    def storeValues(self, settings, key=None):
        if key is None:
            key = self.name
        settings.setValue("Profile:" + key + ":name", QVariant(self.name))
        settings.setValue("Profile:" + key + ":filetype", QVariant(self.filetype))
        settings.setValue("Profile:" + key + ":ext", QVariant(self.ext))
        settings.setValue("Profile:" + key + ":font", QVariant(self.font))
        settings.setValue("Profile:" + key + ":fixedPitch", QVariant(self.fixedPitch))
        settings.setValue("Profile:" + key + ":tabStop", QVariant(self.tabStop))
        settings.setValue("Profile:" + key + ":convertTabs", QVariant(self.convertTabs))
        settings.setValue("Profile:" + key + ":maintainIndent", QVariant(self.maintainIndent))
        settings.setValue("Profile:" + key + ":showLineNumbers", QVariant(self.showLineNumbers))
        settings.setValue("Profile:" + key + ":highlightCurrentLine", QVariant(self.highlightCurrentLine))
        settings.setValue("Profile:" + key + ":highlightLineColor", QVariant(self.highlightLineColor))
        settings.setValue("Profile:" + key + ":wordWrap", QVariant(self.wordWrap))
        settings.setValue("Profile:" + key + ":formatList", QVariant( '|'.join(self.formats.keys()) ))
        for format in self.formats.keys():
            print "saving format", format
            for attr in ['color', 'bold', 'italic']:
                settings.setValue("Profile:%s:format:%s:%s" % (key, format, attr), QVariant(self.formats[format][attr]))
    def loadValues(self, settings, key=None):
        if key is None:
            key = self.name
        self.name = unicode(settings.value("Profile:" + key + ":name").toString())
        self.filetype = unicode(settings.value("Profile:" + key + ":filetype").toString())
        self.ext = settings.value("Profile:" + key + ":ext").toRegExp()
        self.font = QFont(settings.value("Profile:" + key + ":font"))
        self.fixedPitch = settings.value("Profile:" + key + ":fixedPitch").toBool()
        (self.tabStop, Ok) = settings.value("Profile:" + key + ":tabStop").toInt()
        # should raise an exception if not Ok (attribute? type?)
        self.convertTabs = settings.value("Profile:" + key + ":convertTabs").toBool()
        self.maintainIndent = unicode(settings.value("Profile:" + key + ":maintainIndent").toString())
        self.showLineNumbers = settings.value("Profile:" + key + ":showLineNumbers").toBool()
        self.highlightCurrentLine = settings.value("Profile:" + key + ":highlightCurrentLine").toBool()
        self.highlightLineColor = QColor(settings.value("Profile:" + key + ":highlightLineColor"))
        self.wordWrap = unicode(settings.value("Profile:" + key + ":wordWrap").toString())
        formatList = unicode(settings.value("Profile:" + key + ":formatList").toString())
        formatList = formatList.split('|')
        for format in formatList:
            print "loading format", format
            color = QColor(settings.value("Profile:%s:format:%s:color" % (key, format)))
            bold = settings.value("Profile:%s:format:%s:bold" % (key, format)).toBool()
            italic = settings.value("Profile:%s:format:%s:italic" % (key, format)).toBool()
            self.formats[format] = {'color': color, 'bold': bold, 'italic': italic }
    def addRules(self, rule):
        self.highlighter.Rules.append( rule )

class MainWindow(QMainWindow):
    def __init__(self, filename=None, parent=None):
        super(MainWindow, self).__init__(parent)
        initialProfile = 'default'
        self.fontOveride = None
        self.wordWrapMethods = {}
        self.wordWrapMethods['No Wrap'] = {'data': QTextOption.NoWrap, 'hint': 'Text is not wrapped'}
        self.wordWrapMethods['Word Wrap'] = {'data': QTextOption.WordWrap, 'hint': 'Text wraps at word boundaries'}
        #self.wordWrapMethods['Manual Wrap'] = {'data': QTextOption.ManualWrap, 'hint': 'Text is not wrapped'}
        self.wordWrapMethods['Wrap Anywhere'] = {'data': QTextOption.WrapAnywhere, 'hint': 'Text wraps at window edge'}
        self.wordWrapMethods['Flexible Wrap'] = {'data': QTextOption.WrapAtWordBoundaryOrAnywhere, 'hint': 'Wrap at word boundary if possible, otherwise wrap as appropriate—including in the middle of a word'}
        self.helpContentsDlg = None
        self.searchFindDlg = None
        self.searchReplaceDlg = None
        self.toolsPrefsDlg = None
        self.toolsTabDlg = None
        self.macroRepeatDlg = None
        self.macroManageDlg = None
        self.editSetDateFormatDlg = None
        self.storedMacroList = {}
        self.dateFormat = '%Y%m%d %H:%M:%S'
        self.printer = None
        settings = QSettings(__companyName__, __thisAppName__)
        self.font = QFont(settings.value("Font"))
        if self.font is None:
            self.font = QFont("Courier", 10)
        self.font.setFixedPitch(True)
        self.font.setKerning(False)
        self.profiles = {}
        for profileName in ['default', 'pgsTextCode', 'pgsResource', 'python', 'html']:
            self.profiles[profileName] = self.createProfile(profileName)
            self.profiles[profileName].font = self.font
        self.maintainIndentMethods = {}
        self.recentFiles = settings.value("RecentFiles").toStringList()
        self.restoreGeometry(settings.value("Geometry").toByteArray())
        # load date format
        dateFormat = str(settings.value("dateFormat").toString())
        if dateFormat: self.dateFormat = dateFormat
        # load profiles
        profileList = settings.value("ProfileList").toStringList()
        for profile in profileList:
            profile = unicode(profile)
            try:
                self.profiles[profile].loadValues(settings, profile)
            except KeyError:
                print "missing profile:", profile
                self.profiles[profile] = prefsProfile(profile)
        # load macros
        macroList = settings.value("MacroList").toStringList()
        for macro in macroList:
            self.storedMacroList[macro] = []
            (count, Ok) = settings.value("macro:" + macro + ":count").toInt()
            for i in range(0, count):
                (key, Ok) = settings.value("macro:" + macro + ":" + str(i) + ":key").toInt()
                (modifiers, Ok) = settings.value("macro:" + macro + ":" + str(i) + ":modifiers").toInt()
                text = settings.value("macro:" + macro + ":" + str(i) + ":text").toString()
                self.storedMacroList[macro].append( QKeyEvent(QEvent.KeyPress, key, Qt.KeyboardModifiers(modifiers), text) )
        # set up the editor
        self.editor = TextEdit()
        self.editor.app = self
        (tabStop, ok) = settings.value("Tabstop").toInt()
        if ok:
            self.editor.setTabStop(tabStop)
        self.editor.setConvertTabs(settings.value("ConvertTabs").toBool())
        self.editor.setFont(self.font)
        self.editor.setWordWrapMode(self.wordWrapMethods[self.profiles[initialProfile].wordWrap]['data'])
        self.numberBar = NumberBar()
        self.numberBar.setTextEdit(self.editor)
        self.numberBar.visible = self.profiles[initialProfile].showLineNumbers
        # initializing highlighting
        self.highlighter = PythonHighlighter(self.editor.document())
        self.highlighter.setRules(self.profiles[initialProfile].rules)
        self.highlighter.setFormats(self.profiles[initialProfile].formats)
        self.highlighter.useBlocks = self.profiles[initialProfile].useBlocks
        # setting up window
        self.mainFrame = QFrame()
        hbox = QHBoxLayout(self.mainFrame)
        hbox.setSpacing(0)
        hbox.setMargin(0)
        hbox.addWidget(self.numberBar)
        hbox.addWidget(self.editor)
        self.setCentralWidget(self.mainFrame)
        self.editor.installEventFilter(self)
        self.editor.viewport().installEventFilter(self)
        status = self.statusBar()
        status.setSizeGripEnabled(False)
        status.showMessage("Ready", 5000)
        # define actions
        fileNewAction = self.createAction("&New Document", self.fileNew,
                QKeySequence.New, "filenew", "Create a text file")
        fileNewWindowAction = self.createAction("&New Window", self.fileNewWindow,
                QKeySequence.AddTab, "filenewwindow", "Create a new editor window")
        fileOpenAction = self.createAction("&Open...", self.fileOpen,
                QKeySequence.Open, "fileopen",
                "Open an existing text file")
        self.fileSaveAction = self.createAction("&Save", self.fileSave,
                QKeySequence.Save, "filesave", "Save the file")
        self.fileSaveAsAction = self.createAction("Save &As...",
                self.fileSaveAs, "Ctrl+Shift+S", "filesaveas",
                "Save the file using a new name")
        self.fileReloadAction = self.createAction("Reload file",
                self.fileReload, None, "Reload file from disk",
                "Save the file using a new name")
        filePrintAction = self.createAction("Print file",
                self.print_, QKeySequence.Print, None,
                "Print the current document")
        fileCloseAction = self.createAction("Close", self.close,
                QKeySequence.Close, "fileclose", "Close the current window")
        fileQuitAction = self.createAction("&Quit", self.closeApplication,
                "Ctrl+Q", "filequit", "Close the application")
        self.editUndoAction = self.createAction("Undo",
                self.editor.undo, QKeySequence.Undo, "editundo",
                None)
        self.editRedoAction = self.createAction("Redo",
                self.editor.redo, QKeySequence.Redo, "editredo",
                None)
        editSelectAllAction = self.createAction("Select &All",
                self.editor.selectAll, QKeySequence.SelectAll, "editselectAll",
                "Select all text")
        self.editCopyAction = self.createAction("&Copy",
                self.editor.copy, QKeySequence.Copy, "editcopy",
                "Copy text to the clipboard")
        self.editCutAction = self.createAction("Cu&t", self.editor.cut,
                QKeySequence.Cut, "editcut",
                "Cut text to the clipboard")
        self.editPasteAction = self.createAction("&Paste",
                self.editor.paste, QKeySequence.Paste, "editpaste",
                "Paste in the clipboard's text")
        editLowerAction = self.createAction("lowercase",
                self.lowerCase, None, None,
                "Lowercase selected text")
        editUpperAction = self.createAction("UPPERCASE",
                self.upperCase, None, None,
                "Uppercase selected text")
        editCapitalizeAction = self.createAction("Capitalize",
                self.mixedCase, None, None,
                "Capitalize selected text")
        editInsertDateAction = self.createAction("Insert Date",
                self.editInsertDate, None, None,
                "Insert date/time string")
        editSetDateFormatAction = self.createAction("Set Date Format...",
                self.editSetDateFormat, None, None,
                "Set formate for inserting date date/time strings")
        searchFindAction = self.createAction("Find...", self.searchFind, 
                QKeySequence.Find, None, "Open find text dialog")
        searchFindSelectionAction = self.createAction("Find Selection", 
                self.searchFindSelection, "Ctrl+H", None, 
                "Find currently selected text")
        searchFindPrevSelectionAction = self.createAction("Find Prev Selection", 
                self.searchFindPrevSelection, "Ctrl+Shift+H", None, 
                "Find currently selected text searching backwards")
        searchFindNextAction = self.createAction("Find Next", 
                self.searchFindNext, QKeySequence.FindNext, None, "Find next instance of text or pattern")
        searchFindPrevAction = self.createAction("Find Previous", 
                self.searchFindPrev, "Ctrl+Shift+G", None, "Find previous instance of text or pattern")
        searchReplaceAction = self.createAction("Replace...", 
                self.searchReplace, QKeySequence.Replace, None, "Open find/replace dialog")
        searchReplaceNextAction = self.createAction("Replace Next",
                self.searchReplaceNext, "Ctrl+T",
                None, "Replace current selection and find again")
        # Ctrl+T for searchReplaceNextAction collides with Ctrl+T for OSX (new window)
        if platform.system() == 'Darwin':
            searchReplaceNextAction = self.createAction("Replace Next",
                    self.searchReplaceNext, None,
                    None, "Replace current selection and find again")
        searchReplaceAllAction = self.createAction("Replace All", 
                self.searchReplaceAll, None, None, 
                "Replace all instances of current find selection in text with current replace selection")
        searchGotoLineAction = self.createAction("Goto Line",
                self.searchGotoLine, "Ctrl+L", None,
                "Goto line (paragraph) number")
        self.macroStartAction = self.createAction("Record", self.macroStart,
                "Alt+K", None, "Start recording key strokes")
        self.macroStopAction = self.createAction("Finish", self.macroStop,
                "Alt+K", None, "Stop recording key strokes and remember as macro")
        self.macroCancelAction = self.createAction("Cancel", self.macroCancel,
                "Alt+Shift+K", None, "Stop recording key strokes and forget them")
        self.macroReplayAction = self.createAction("Replay", self.macroReplay,
                "Ctrl+K", None, "Replay current macro")
        self.macroRepeatAction = self.createAction("Repeat...", 
                self.macroRepeat, None, None, 
                "Open dialog for repeating current macro")
        self.macroManageAction = self.createAction("Manage...", 
                self.macroManage, None, None, 
                "Open dialog for managing macros")
        toolsSortAction = self.createAction("Sort", self.toolsSort, None, 
                None, "Sort selected text")
        toolsPrefsAction = self.createAction("Preferences...", self.toolsPrefs, 
                None, None, "Open preferences dialog")
        toolsFontAction = self.createAction("Select Font...", self.toolsFont,
                None, None, "Open font selection dialog")
        toolsResetFontAction = self.createAction("Reset Font", self.toolsResetFont,
                None, None, "Reset font to style setting")
        toolsTabAction = self.createAction("Set Tabstop...", self.toolsTab,
                None, None, "Open dialog to set conversion of tabs to spaces")
        self.toolsShowLineNumberAction = self.createAction("Show Line Numbers", self.toolsShowLineNumber,
                None, None, "Toggle the display of line (paragraph) numbers", True)
        self.toolsHighlightCurrentLineAction = self.createAction("Highlight Current Line", self.toolsHighlightCurrentLine,
                None, None, "Toggle the highlighting of the current line", True)
        self.toolsHighlightLineColorAction = self.createAction("Highlight Color...", self.toolsHighlightLineColor,
                None, None, "Open dialog to set the background color to use when highlighting the current line")
        helpAboutAction = self.createAction("About...", self.helpAbout, None,
                None, "About " + __thisAppName__)
        helpContentsAction = self.createAction("Contents...", self.helpContents,
                QKeySequence.HelpContents, None, "View help file")
        self.fileMenu = self.menuBar().addMenu("&File")
        self.fileMenuActions = (fileNewAction, fileNewWindowAction, fileOpenAction,
                self.fileSaveAction, self.fileSaveAsAction, self.fileReloadAction, None,
                filePrintAction, None, fileCloseAction, fileQuitAction)
        editMenu = self.menuBar().addMenu("&Edit")
        self.addActions(editMenu, (self.editUndoAction, self.editRedoAction,
                None, 
                self.editCopyAction,
                self.editCutAction, self.editPasteAction, None,
                editSelectAllAction, None, editLowerAction,
                editUpperAction, editCapitalizeAction, None,
                editInsertDateAction, editSetDateFormatAction))
        searchMenu = self.menuBar().addMenu("&Search")
        self.addActions(searchMenu, (searchFindAction, searchFindSelectionAction, 
                searchFindPrevSelectionAction, searchFindNextAction, 
                searchFindPrevAction, searchReplaceAction, 
                searchReplaceNextAction, searchReplaceAllAction, None, searchGotoLineAction))
        macroMenu = self.menuBar().addMenu("Ma&cro")
        self.addActions(macroMenu, (self.macroStartAction, self.macroStopAction, 
                self.macroCancelAction, self.macroReplayAction, self.macroRepeatAction, 
                self.macroManageAction))
        toolsMenu = self.menuBar().addMenu("&Tools")
        self.addActions(toolsMenu, (toolsSortAction, None, toolsPrefsAction, toolsFontAction, toolsResetFontAction, toolsTabAction,
                self.toolsShowLineNumberAction, self.toolsHighlightCurrentLineAction, self.toolsHighlightLineColorAction))
        maintainIndentToolsMenu = toolsMenu.addMenu("Automatic Indentation")
        self.maintainIndentGroup = QActionGroup(self)
        for method in ['None', 'Normal', 'Smart']:
            action = QAction(method, self)
            action.setData(method)
            action.setCheckable(True)
            action.setToolTip("Set maintain indentation method to %s" % method)
            action.setStatusTip("Set maintain indentation method to %s" % method)
            self.connect(action, SIGNAL("triggered()"), self.setMaintainIndent)
            maintainIndentToolsMenu.addAction(action)
            self.maintainIndentGroup.addAction(action)
            self.maintainIndentMethods[method] = action
            if method == self.profiles[initialProfile].maintainIndent: action.setChecked(True)
        wordWrapToolsMenu = toolsMenu.addMenu("Word Wrap")
        self.wordWrapGroup = QActionGroup(self)
        for method in self.wordWrapMethods:
            action = QAction(method, self)
            action.setData(method)
            action.setCheckable(True)
            action.setToolTip(self.wordWrapMethods[method]['hint'])
            action.setStatusTip(self.wordWrapMethods[method]['hint'])
            self.connect(action, SIGNAL("triggered()"), self.setWordWrap)
            wordWrapToolsMenu.addAction(action)
            self.wordWrapGroup.addAction(action)
            self.wordWrapMethods[method]['action'] = action
            if method == self.profiles[initialProfile].wordWrap: action.setChecked(True)
        profileToolsMenu = toolsMenu.addMenu("Select Profile")
        self.profileGroup = QActionGroup(self)
        for profileName in self.profiles:
            profName = self.profiles[profileName].name
            action = QAction(profName, self)
            action.setData(QVariant(profileName))
            action.setCheckable(True)
            action.setToolTip("Set profile to %s" % profName)
            action.setStatusTip("Set profile to %s" % profName)
            self.connect(action, SIGNAL("triggered()"), self.setProfile)
            profileToolsMenu.addAction(action)
            self.profileGroup.addAction(action)
            self.profiles[profileName].action = action
            if profileName == initialProfile: action.setChecked(True)
        self.toolsShowLineNumberAction.setChecked( self.profiles[initialProfile].showLineNumbers )
        self.toolsHighlightCurrentLineAction.setChecked( self.profiles[initialProfile].highlightCurrentLine )
        helpMenu = self.menuBar().addMenu("&Help")
        self.addActions(helpMenu, (helpAboutAction, helpContentsAction))
        fileToolbar = self.addToolBar("File")
        fileToolbar.setObjectName("FileToolBar")
        self.addActions(fileToolbar, (fileNewAction, fileOpenAction,
                                      self.fileSaveAction))
        editToolbar = self.addToolBar("Edit")
        editToolbar.setObjectName("EditToolBar")
        self.addActions(editToolbar, (self.editCopyAction,
                self.editCutAction, self.editPasteAction))
        self.connect(self.fileMenu, SIGNAL("aboutToShow()"), self.updateFileMenu)
        self.connect(self.editor,
                SIGNAL("selectionChanged()"), self.updateUi)
        self.connect(self.editor.document(),
                SIGNAL("modificationChanged(bool)"), self.updateUi)
        self.connect(QApplication.clipboard(),
                SIGNAL("dataChanged()"), self.updateUi)
        self.connect(self.editor.document(), SIGNAL("undoAvailable(bool)"), self.updateUndo)
        self.connect(self.editor.document(), SIGNAL("redoAvailable(bool)"), self.updateRedo)
        self.connect(self.editor, SIGNAL("cursorPositionChanged()"), self.updateHighlight)
        self.setWindowTitle(__thisAppName__)
        self.filename = filename
        self.updateFileMenu()
        if self.filename is not None:
            QTimer.singleShot(0, self.loadInitialFile )
        else: self.applyProfile(initialProfile)
        self.editUndoAction.setEnabled(False)
        self.editRedoAction.setEnabled(False)
        self.updateUi()
    def eventFilter(self, object, event):
        # update the line numbers for all events on the text edit and the view port.
        # This is easier than connecting all necessary signals
        if object in (self.editor, self.editor.viewport()):
            self.numberBar.update()
            return False
        return QFrame.eventFilter(object, event)
    def loadInitialFile(self):
        if self.filename and QFile.exists(self.filename):
            self.loadFile(self.filename)
        else:
            settings = QSettings(__companyName__, __thisAppName__)
            fname = unicode(settings.value("LastFile").toString())
            if fname and QFile.exists(fname):
                self.loadFile(fname)
    def updateUi(self, arg=None):
        self.fileSaveAction.setEnabled(
                self.editor.document().isModified())
        enable = not self.editor.document().isEmpty()
        self.fileSaveAsAction.setEnabled(enable)
        enable = self.editor.textCursor().hasSelection()
        self.editCopyAction.setEnabled(enable)
        self.editCutAction.setEnabled(enable)
        self.editPasteAction.setEnabled(self.editor.canPaste())
        self.macroStartAction.setEnabled( not self.editor.macroRecord )
        self.macroStopAction.setEnabled(self.editor.macroRecord)
        self.macroCancelAction.setEnabled(self.editor.macroRecord)
        self.macroReplayAction.setEnabled(bool(self.editor.macroKeyPresses))
    def updateHighlight(self, force=False):
        # force disabling of this capability due to it changing the modified
        # state of the application. Enable after implementing own undo/redo
        # framework that can exclude this activity
        return
        cursor = self.editor.textCursor()
        block = cursor.block()
        if self.editor.highlightCurrentLine:
            if block == self.editor.lastBlock and not force: return
            if self.editor.lastBlock:
                blockFormat = self.editor.lastBlock.blockFormat()
                blockFormat.clearBackground()
                self.editor.lastCursor.setBlockFormat(blockFormat)
            blockFormat = block.blockFormat()
            blockFormat.setBackground(self.editor.highlightLineColor)
            cursor.setBlockFormat(blockFormat)
            self.editor.lastBlock = block
            self.editor.lastCursor = cursor
        else:
            if self.editor.lastBlock:
                blockFormat = self.editor.lastBlock.blockFormat()
                blockFormat.clearBackground()
                self.editor.lastCursor.setBlockFormat(blockFormat)
    def updateUndo(self, available):
        self.editUndoAction.setEnabled(available)
    def updateRedo(self, available):
        self.editRedoAction.setEnabled(available)
    def createAction(self, text, slot=None, shortcut=None, icon=None,
                     tip=None, checkable=False, signal="triggered()"):
        action = QAction(text, self)
        if icon is not None:
            action.setIcon(QIcon(":/%s.png" % icon))
        if shortcut is not None:
            action.setShortcut(shortcut)
        if tip is not None:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if slot is not None:
            self.connect(action, SIGNAL(signal), slot)
        if checkable:
            action.setCheckable(True)
        return action
    def addActions(self, target, actions):
        for action in actions:
            if action is None:
                target.addSeparator()
            else:
                target.addAction(action)
    def closeEvent(self, event):
        if self.okToContinue():
            settings = QSettings(__companyName__, __thisAppName__)
            filename = QVariant(QString(self.filename)) \
                    if self.filename is not None else QVariant()
            settings.setValue("LastFile", filename)
            recentFiles = QVariant(self.recentFiles) \
                    if self.recentFiles else QVariant()
            settings.setValue("RecentFiles", recentFiles)
            settings.setValue("Geometry", QVariant(self.saveGeometry()))
            settings.setValue("MainWindow/State", QVariant(self.saveState()))
            settings.setValue("Font", QVariant(self.editor.font()))
            settings.setValue("Tabstop", QVariant(self.editor.getTabStop()))
            settings.setValue("ConvertTabs", QVariant(self.editor.getConvertTabs()))
            settings.setValue("ShowLineNumbers", QVariant(self.numberBar.visible))
            settings.setValue("HighlightLine", QVariant(self.editor.highlightCurrentLine))
            settings.setValue("HighlightColor", QVariant(self.editor.highlightLineColor))
            
            settings.setValue("dateFormat", QVariant(self.dateFormat) )
            
            settings.setValue("ProfileList", QVariant( self.profiles.keys() ))
            for profileKey in self.profiles:
                self.profiles[profileKey].storeValues(settings, profileKey)
            settings.setValue("MacroList", QVariant( self.storedMacroList.keys() ))
            for macro in self.storedMacroList:
                settings.setValue("macro:" + macro + ":count", len(self.storedMacroList[macro]))
                i = 0
                for event in self.storedMacroList[macro]:
                    settings.setValue("macro:" + macro + ":" + str(i) + ":key", QVariant(event.key()))
                    settings.setValue("macro:" + macro + ":" + str(i) + ":modifiers", QVariant(event.modifiers()))
                    settings.setValue("macro:" + macro + ":" + str(i) + ":text", QVariant(event.text()))
                    i += 1
        else:
            event.ignore()
    def okToContinue(self):
        if self.editor.document().isModified():
            reply = QMessageBox.question(self,
                            __thisAppName__ + " - Unsaved Changes",
                            "Save unsaved changes?",
                            QMessageBox.Yes|QMessageBox.No|
                            QMessageBox.Cancel)
            if reply == QMessageBox.Cancel:
                return False
            elif reply == QMessageBox.Yes:
                self.fileSave()
        return True
    def applyProfile(self, profileName):
        self.highlighter.setRules(self.profiles[profileName].rules)
        self.highlighter.setFormats(self.profiles[profileName].formats)
        self.highlighter.useBlocks = self.profiles[profileName].useBlocks
        self.numberBar.visible = self.profiles[profileName].showLineNumbers
        self.toolsShowLineNumberAction.setChecked( self.profiles[profileName].showLineNumbers )
        self.toolsHighlightCurrentLineAction.setChecked( self.profiles[profileName].highlightCurrentLine )
        font = self.profiles[profileName].font
        if font is None:
            font = self.font
        if self.fontOveride:
            font = self.fontOveride
        font.setFixedPitch(self.profiles[profileName].fixedPitch)
        font.setKerning(not self.profiles[profileName].fixedPitch)
        self.editor.setFont(font)
        self.editor.setTabStop(self.profiles[profileName].tabStop)
        self.editor.convertTabs = self.profiles[profileName].convertTabs
        self.editor.maintainIndent = self.profiles[profileName].maintainIndent
        self.profiles[profileName].action.setChecked(True)
        self.maintainIndentMethods[self.profiles[profileName].maintainIndent].setChecked(True)
        self.editor.setWordWrapMode(
            self.wordWrapMethods[self.profiles[profileName].wordWrap]['data'])
        self.wordWrapMethods[self.profiles[profileName].wordWrap]['action'].setChecked(True)
        self.editor.highlightCurrentLine = self.profiles[profileName].highlightCurrentLine
        self.editor.highlightLineColor = self.profiles[profileName].highlightLineColor
        self.highlighter.rehighlight()
        
    def closeApplication(self):
        # FIXME: how to make this all windows in application?
        for window in all_windows:
            window.close()
    def print_(self):
        if self.printer is None:
            self.printer = QPrinter()
        dialog = QPrintDialog(self.printer)
        if not dialog.exec_():
            return
        self.editor.print_(self.printer)

    def fileNew(self):
        if not self.okToContinue():
            return
        document = self.editor.document()
        document.clear()
        document.setModified(False)
        self.filename = None
        self.setWindowTitle(__thisAppName__ + " - Unnamed")
        self.applyProfile( 'default' )
        self.updateUi()
    def fileNewWindow(self):
        form = MainWindow(None)
        form.show()
        all_windows.append(form)
    def fileOpen(self):
        if not self.okToContinue():
            return
        dir = os.path.dirname(str(self.filename)) \
            if self.filename is not None else "."
        formatList = []
        for profileName in self.profiles:
            formatList.append( self.profiles[profileName].filetype )
        availFormats = ";;".join(formatList)
        fname = unicode(QFileDialog.getOpenFileName(self,
                            __thisAppName__ + " - Choose File", dir,
                            availFormats))
        if fname:
            self.loadFile(fname)
    def fileReload(self):
        if not self.okToContinue():
            return
        self.loadFile(self.filename)
    def loadFile(self, fname=None):
        if fname is None:
            action = self.sender()
            if isinstance(action, QAction):
                fname = unicode(action.data().toString())
                if not self.okToContinue():
                    return
            else:
                return
        if fname is None: return
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        # select profile based on extension
        for profileName in self.profiles:
            if self.profiles[profileName].ext.indexIn(fname) >= 0:
                self.applyProfile( profileName )
                break
        fh = None
        try:
            self.filename = None
            fh = QFile(fname)
            if not fh.open(QIODevice.ReadOnly):
                raise IOError, unicode(fh.errorString())
            stream = QTextStream(fh)
            stream.setCodec("UTF-8")
            self.editor.setPlainText(stream.readAll())
            self.editor.document().setModified(False)
        except (IOError, OSError), e:
            QMessageBox.warning(self, __thisAppName__ + " -- Load Error",
                    "Failed to load %s: %s" % (self.filename, e))
        else:
            self.addRecentFile(fname)
            self.setWindowTitle(__thisAppName__ + " - %s" % \
                    QFileInfo(fname).fileName())
            self.filename = fname
        finally:
            if fh is not None:
                fh.close()
            QApplication.restoreOverrideCursor()
    def fileSave(self):
        if self.filename is None:
            return self.fileSaveAs()
        fh = None
        try:
            fh = QFile(self.filename)
            if not fh.open(QIODevice.WriteOnly):
                raise IOError, unicode(fh.errorString())
            stream = QTextStream(fh)
            stream.setCodec("UTF-8")
            stream << self.editor.toPlainText()
            self.editor.document().setModified(False)
        except (IOError, OSError), e:
            QMessageBox.warning(self, __thisAppName__ + " -- Save Error",
                    "Failed to save %s: %s" % (self.filename, e))
        finally:
            if fh is not None:
                fh.close()
        return True
    def fileSaveAs(self):
        filename = self.filename if self.filename is not None else "."
        filename = QFileDialog.getSaveFileName(self,
                        __thisAppName__ + " -- Save File As",
                        filename, "PageStream text files (*.txt)")
        if not filename.isEmpty():
            self.filename = filename
            self.setWindowTitle(__thisAppName__ + " - %s" % \
                    QFileInfo(self.filename).fileName())
            self.addRecentFile(self.filename)
            return self.fileSave()
        return False
    def updateFileMenu(self):
        self.fileMenu.clear()
        self.addActions(self.fileMenu, self.fileMenuActions[:-2])
        current = QString(self.filename) \
                if self.filename is not None else None
        recentFiles = []
        if current is not None:
            recentFiles.append(current)
        for fname in self.recentFiles:
            if fname != current and QFile.exists(fname):
                recentFiles.append(fname)
        if recentFiles:
            self.fileMenu.addSeparator()
            for i, fname in enumerate(recentFiles):
                action = self.createAction("&%d %s" % (i+1, QFileInfo(fname).fileName()), 
                        self.loadFile, None,
                        QIcon(":/icon.png"), "Load %s" % fname)
                action.setData(QVariant(fname))
                self.connect(action, SIGNAL("triggered()"), self.loadFile)
                self.fileMenu.addAction(action)
        self.fileMenu.addSeparator()
        self.fileMenu.addActions(self.fileMenuActions[-2:])
    
    def addRecentFile(self, fname):
        if fname is None:
            return
        self.recentFiles.prepend(QString(fname))
        if self.recentFiles.removeDuplicates() < 1:
            while self.recentFiles.count() > 9:
                self.recentFiles.takeLast()
        self.updateFileMenu()

    def updateProfile(self):
        index = self.toolsPrefsDlg.profileComboBox.currentIndex()
        selectedProfile = unicode(self.toolsPrefsDlg.profileComboBox.itemData(index).toString())
        self.profiles[selectedProfile].tabStop = self.toolsPrefsDlg.tabstopSpinBox.value()
        self.profiles[selectedProfile].convertTabs = self.toolsPrefsDlg.tabToSpaceCheckBox.checkState()
        self.profiles[selectedProfile].showLineNumbers = self.toolsPrefsDlg.showLineNumbersCheckBox.checkState()
        self.profiles[selectedProfile].maintainIndent = unicode(self.toolsPrefsDlg.maintainIndentComboBox.currentText())
        self.profiles[selectedProfile].wordWrap = unicode(self.toolsPrefsDlg.wordWrapComboBox.currentText())
        font = self.toolsPrefsDlg.fontComboBox.currentFont()
        font.setPointSize( self.toolsPrefsDlg.fontSizeSpinBox.value() )
        self.profiles[selectedProfile].font = font
        self.profiles[selectedProfile].highlightCurrentLine = self.toolsPrefsDlg.highlightCheckBox.checkState()
        self.profiles[selectedProfile].highlightLineColor = self.toolsPrefsDlg.highlightLineColor
        # pardon the ugly code, PyQt does not provide a pythonized version of QtListWidget
        row = 0
        listWidgetItem = self.toolsPrefsDlg.styleListWidget.item(row)
        while listWidgetItem is not None:
            color = QColor(listWidgetItem.data(32))
            bold = listWidgetItem.data(33).toBool()
            italic = listWidgetItem.data(34).toBool()
            formatName = unicode(listWidgetItem.text())
            self.profiles[selectedProfile].formats[formatName]['color'] = color
            self.profiles[selectedProfile].formats[formatName]['bold'] = bold
            self.profiles[selectedProfile].formats[formatName]['italic'] = italic
            row += 1
            listWidgetItem = self.toolsPrefsDlg.styleListWidget.item(row)
        self.toolsPrefsDlg.hide()
        if unicode(self.profileGroup.checkedAction().data().toString()) == selectedProfile:
            self.applyProfile(selectedProfile)
        
    def restoreDefaultProfile(self):
        index = self.toolsPrefsDlg.profileComboBox.currentIndex()
        selectedProfile = unicode(self.toolsPrefsDlg.profileComboBox.itemData(index).toString())
        action = self.profiles[selectedProfile].action
        self.profiles[selectedProfile] = self.createProfile(selectedProfile)
        self.profiles[selectedProfile].action = action
        if self.profiles[selectedProfile].font is None:
            self.profiles[selectedProfile].font = self.editor.font()
        if unicode(self.profileGroup.checkedAction().data().toString()) == selectedProfile:
            self.applyProfile(selectedProfile)
            # make sure the toolsPrefsDlg is updated back to the defaults
            self.toolsPrefsDlg.indexChanged(index)
        
    def setProfile(self):
        action = self.sender()
        if isinstance(action, QAction):
            profileName = unicode(action.data().toString())
            self.applyProfile( profileName )
    
    def setMaintainIndent(self):
        action = self.sender()
        if isinstance(action, QAction):
            method = unicode(action.data().toString())
            self.editor.maintainIndent = method

    def setWordWrap(self):
        action = self.sender()
        if isinstance(action, QAction):
            method = unicode(action.data().toString())
            self.editor.setWordWrapMode( self.wordWrapMethods[method]['data'] )

    def toolsSetFont(self):
        font = self.profiles[ unicode(self.profileGroup.checkedAction().data().toString()) ].font
        if font is None:
            font = self.font
        (font, ok) = QFontDialog.getFont(font, self)
        if ok:
            self.fontOveride = font
            self.editor.setFont(font)
        
    def toolsSort(self):
        if self.editor.textCursor().hasSelection():
            # apply to selection
            pass
        else:
            # apply to entire text
            pass
    def toolsFont(self):
        profileName = unicode(self.profileGroup.checkedAction().data().toString())
        font = self.profiles[ profileName ].font
        if font is None:
            font = self.font
        (font, ok) = QFontDialog.getFont(font, self)
        if ok:
            self.fontOveride = font
            self.font.setFixedPitch( self.profiles[ profileName ].fixedPitch )
            self.font.setKerning( not self.profiles[ profileName ].fixedPitch )
            self.editor.setFont(font)
    def toolsResetFont(self):
        self.fontOveride = None
        profileName = unicode(self.profileGroup.checkedAction().data().toString())
        font = self.profiles[ profileName ].font
        if font is None:
            font = self.font
        self.font.setFixedPitch( self.profiles[ profileName ].fixedPitch )
        self.font.setKerning( not self.profiles[ profileName ].fixedPitch )
        self.editor.setFont(font)
    def toolsTab(self):
        if not self.toolsTabDlg:
            self.toolsTabDlg = ToolsTabDlg(self.storeTabstop, self)
            self.toolsTabDlg.setWindowTitle("Set tab options")
        self.toolsTabDlg.setTabstop( self.editor.getTabStop() )
        self.toolsTabDlg.setConvertTabs( self.editor.getConvertTabs() )
        self.toolsTabDlg.show()
    def storeTabstop(self):
        self.editor.setTabStop( self.toolsTabDlg.tabstopSpinBox.value() )
        self.editor.setConvertTabs( self.toolsTabDlg.tabToSpaceCheckBox.checkState() )
        self.toolsTabDlg.hide()
    # could/should do this via the notifier signal of toggled
    def toolsShowLineNumber(self):
        action = self.sender()
        self.numberBar.visible = action.isChecked()
    def toolsHighlightCurrentLine(self):
        action = self.sender()
        self.editor.highlightCurrentLine = action.isChecked()
        self.updateHighlight()
    def toolsHighlightLineColor(self):
        color = QColorDialog().getColor( self.editor.highlightLineColor )
        self.editor.highlightLineColor = color
        self.updateHighlight(True)
    def toolsPrefs(self):
        if not self.toolsPrefsDlg:
            self.toolsPrefsDlg = ToolsPrefsDlg(self.updateProfile, self.restoreDefaultProfile, self)
            self.toolsPrefsDlg.setWindowTitle('Preferences')
        self.toolsPrefsDlg.profileComboBox.setCurrentIndex(
            self.toolsPrefsDlg.profileComboBox.findData(
            self.profileGroup.checkedAction().data().toString()))
        self.toolsPrefsDlg.show()

    def lowerCase(self):
        textCursor = self.editor.textCursor()
        textCursor.insertText( textCursor.selectedText().toLower() )

    def upperCase(self):
        textCursor = self.editor.textCursor()
        textCursor.insertText( textCursor.selectedText().toUpper() )

    def mixedCase(self):
        textCursor = self.editor.textCursor()
        textCursor.insertText( unicode(textCursor.selectedText()).capitalize() )

    def editInsertDate(self):
        textCursor = self.editor.textCursor()
        textCursor.insertText( datetime.datetime.today().strftime(self.dateFormat) )
    
    def editSetDateFormat(self):
        if not self.editSetDateFormatDlg:
            self.editSetDateFormatDlg = EditSetDateFormatDlg(self.getDateFormat, self.setDateFormat)
            self.editSetDateFormatDlg.setWindowTitle('Set Date Format')
        self.editSetDateFormatDlg.show()
    def getDateFormat(self):
        return self.dateFormat
    def setDateFormat(self, dateFormat):
        self.dateFormat = dateFormat

    def searchGotoLine(self):
        cursor = self.editor.textCursor()
        lineNumber, ok = QInputDialog.getInteger(self, "Goto Line Number", "Line:",
            cursor.blockNumber() + 1, 1)
        if ok:
            cursor.movePosition(cursor.Start, cursor.MoveAnchor)
            cursor.movePosition(cursor.NextBlock, cursor.MoveAnchor, lineNumber - 1)
            self.editor.setTextCursor(cursor)
    def searchFind(self):
        if self.searchFindDlg:
            self.searchFindDlg.resetText()
        else:
            self.searchFindDlg = SearchFindDlg(self.searchFindNext, self)
            self.searchFindDlg.setWindowTitle('Find text')
        self.searchFindDlg.show()
    def searchFindPrev(self):
        self.searchFindNext(backwards=True)
    def searchFindNext(self, searchText=None, backwards=False):
        if searchText is None:
            searchText = self.searchFindDlg.findField.text()
        if not searchText:
            return
        caseSensitive = self.searchFindDlg.caseSensitive.isChecked()
        findBackwards = self.searchFindDlg.findBackward.isChecked()
        matchWholeWord = True if self.searchFindDlg.matchComboBox.currentIndex() == 1 else False
        if backwards:
            findBackwards = True
        searchTerm = searchText
        if self.searchFindDlg.matchComboBox.currentIndex() == 2:
            # indicates a regular expression
            searchTerm = QRegExp(r"%s" % searchText)
        found = self.findText( searchTerm, caseSensitive, findBackwards, matchWholeWord )
        if found:
            self.searchFindDlg.hide()
        else:
            reply = QMessageBox.question(self, "Text not found",
                            "Search pattern '%s' not found\nWrap search?" % searchText,
                            QMessageBox.Yes|QMessageBox.No|
                            QMessageBox.Cancel)
            if reply == QMessageBox.Cancel:
                self.searchFindDlg.hide()
            elif reply == QMessageBox.Yes:
                textCursor = self.editor.textCursor()
                if findBackwards:
                    textCursor.setPosition(self.editor.document().characterCount() - 1)
                else:
                    textCursor.setPosition(0)
                self.editor.setTextCursor(textCursor)
                self.searchFindNext(searchText, backwards)
    def findText(self, searchTerm=None, casesensitive=False, backwards=False, wholeword=False):
        if not searchTerm:
            return False
        caseSensitive = QTextDocument.FindCaseSensitively \
                if casesensitive else QTextDocument.FindFlags(0)
        findBackwards = QTextDocument.FindBackward \
                if backwards else QTextDocument.FindFlags(0)
        matchWholeWord = QTextDocument.FindWholeWords \
                if wholeword else QTextDocument.FindFlags(0)
        if isinstance(searchTerm, QRegExp) and not caseSensitive:
            searchTerm.setCaseSensitivity(Qt.CaseInsensitive)
        document = self.editor.document()
        oldCursor = self.editor.textCursor()
        newCursor = document.find(searchTerm, oldCursor, caseSensitive | findBackwards | matchWholeWord)
        # ugly -- the python equivalent of a null pointer is None
        if newCursor.isNull():
            return False
        else:
            self.editor.setTextCursor(newCursor)
            return True
        
    def searchFindSelection(self, backwards=False):
        if not self.searchFindDlg:
            self.searchFindDlg = SearchFindDlg(self.searchFindNext, self)
            self.searchFindDlg.setWindowTitle('Find text')
        searchText = self.editor.textCursor().selectedText()
        if searchText != self.searchFindDlg.findField.text():
            self.searchFindDlg.findField.setText( searchText )
        self.searchFindNext( searchText, backwards )
    def searchFindPrevSelection(self):
        self.searchFindSelection(backwards=True)

    def searchReplace(self):
        if self.searchReplaceDlg:
            self.searchReplaceDlg.resetText()
        else:
            self.searchReplaceDlg = SearchReplaceDlg(self.searchReplaceNext, self)
            self.searchReplaceDlg.setWindowTitle('Replace text')
        self.searchReplaceDlg.show()
    def searchReplaceNext(self, searchText=None, replaceText=None, replaceAll=False):
        if searchText is None:
            searchText = self.searchReplaceDlg.findField.text()
        if replaceText is None:
            replaceText = self.searchReplaceDlg.replaceField.text()
        if not searchText:
            return
        caseSensitive = self.searchReplaceDlg.caseSensitive.isChecked()
        findBackwards = self.searchReplaceDlg.findBackward.isChecked()
        matchWholeWord = True if self.searchReplaceDlg.matchComboBox.currentIndex() == 1 else False
        
        searchTerm = searchText
        if self.searchReplaceDlg.matchComboBox.currentIndex() == 2:
            # indicates a regular expression
            searchTerm = QRegExp(r"%s" % searchText)
        self.searchReplaceDlg.hide()
        while self.findText(searchTerm, caseSensitive, findBackwards, matchWholeWord):
            textCursor = self.editor.textCursor()
            textCursor.insertText( replaceText )
            if not replaceAll: break
        else:
            reply = QMessageBox.question(self, "Text not found",
                            "Search pattern '%s' not found\nWrap search?" % searchText,
                            QMessageBox.Yes|QMessageBox.No|
                            QMessageBox.Cancel)
            if reply == QMessageBox.Cancel:
                self.searchReplaceDlg.hide()
            elif reply == QMessageBox.Yes:
                textCursor = self.editor.textCursor()
                if findBackwards:
                    textCursor.setPosition(textCursor.End)
                else:
                    textCursor.setPosition(textCursor.Start)
                self.editor.setTextCursor(textCursor)
                self.searchReplaceNext()

    def searchReplaceAll(self):
        if self.searchReplaceDlg:
            self.searchReplaceNext(replaceAll=True)
        else:
          self.searchReplace()

    def macroStart(self):
        self.editor.macroKeyPresses = []
        self.editor.macroRecord = True
        self.updateUi()
    def macroStop(self):
        self.editor.macroRecord = False
        self.updateUi()
    def macroCancel(self):
        self.editor.macroRecord = False
        self.editor.macroKeyPresses = []
        self.updateUi()
    def macroReplay(self):
        for event in self.editor.macroKeyPresses:
            self.editor.event(event)
    def macroRepeat(self):
        # replays macro 'x' times where x is solicited from the user
        if not self.macroRepeatDlg:
            self.macroRepeatDlg = MacroRepeatDlg(self.macroPlay, self)
            self.macroRepeatDlg.setWindowTitle('Repeat Macro')
        self.macroRepeatDlg.show()
    def macroPlay(self):
        try:
            for event in self.storedMacroList[self.macroRepeatDlg.macroComboBox.currentText()]:
                self.editor.event(event)
        except AttributeError:
            self.macroReplay
        self.macroRepeatDlg.hide()

    def macroManage(self):
        # opens a dialog that allows setting current macro, saving and deleting macros, running any stored macro directly for 'n' times
        if not self.macroManageDlg:
            self.macroManageDlg = MacroManageDlg(self)
            self.macroManageDlg.setWindowTitle("Manage Macros")
        self.macroManageDlg.backStore()
        self.macroManageDlg.show()
        
    def toolsSort(self):
        # split text on paragraph breaks
        separator = u'\u2029'
        selectedText = self.editor.textCursor().selectedText()
        stringList = selectedText.split(separator)
        stringList.sort()
        self.editor.textCursor().insertText(stringList.join(separator))

    def helpAbout(self):
        QMessageBox.about(self, "About " + __thisAppName__,
        "<b>" + __thisAppName__ + "</b>" +
        """ v %s
            <p>Copyright &copy; 2010, 2011 Tim Doty</p>
            <p>Derived from a code example in <i>Rapid GUI Programming with Python
            and Qt</i> by Mark Summerfield and published by Prentice Hall. Syntax
            highlighting for python code taken from <i>sandbox.pyw</i> also by
            Mark Summerfield.</p>
            <p>This application is a simple text editor that provides syntax
            highlighting for PageStream text codes with some support for Python and
            HTML.</p>
            <p>Python %s - Qt %s - PyQt %s on %s
        """ % (
            __version__, platform.python_version(), QT_VERSION_STR,
            PYQT_VERSION_STR, platform.system()))
    def helpContents(self):
        if not self.helpContentsDlg:
            self.helpContentsDlg = HelpContentsDlg(__thisHelpFileName__, self)
            self.helpContentsDlg.setWindowTitle(__thisAppName__ + ' Help')
        self.helpContentsDlg.show()

    def createProfile(self, profileName):
        """
            Used to create pre-defined profiles. If no profile name is specified
            or if it is not one of the pre-defined profiles then the default
            profile is used instead. Should probably throw an exception instead.
        """
        if profileName == 'pgsTextCode':
            return self.pgsTextCodeProfile()
        elif profileName == 'pgsResource':
            return self.pgsResourceProfile()
        elif profileName == 'python':
            return self.pythonProfile()
        elif profileName == 'html':
            return self.htmlProfile()
        else: return self.defaultProfile()
        
    def pgsTextCodeProfile(self):
        tProfile = prefsProfile('PgS TextCode', "PageStream text files (*.txt)", QRegExp(r"\.te?xt$"))
        tProfile.setFormatAttr('normal', '#000000', False, False)
        tProfile.setFormatAttr('typestyle', '#008000', True, False)
        tProfile.setFormatAttr('characterattributes', '#800000', True, False)
        tProfile.setFormatAttr('styletags', '#800080', True, False)
        tProfile.setFormatAttr('paragraphformats', '#808000', True, False)
        tProfile.setFormatAttr('specialcharacters', '#4080A0', True, False)
        tProfile.setFormatAttr('metatags', '#000000', True, False)

        tProfile.rules.append((QRegExp(r"@[PC][\w ]*:"), "styletags"))
        tProfile.rules.append((QRegExp(
                r"<[BIUORSM\+\-$]>"), "typestyle"))
        tProfile.rules.append((QRegExp(
                r"<f\"[\w\s]+\">"
                r"|<[zhktTl]-?[\d\.]+(pt|%|\"|cm)?>"
                r"|<L[arfg]>"
                r"|<l[arfg](,\d+%)?>"
                r"|<C(\"\w+\"|c\d+m\d+y\d+k\d+),\d+(,(\"\w+\"|c\d+m\d+y\d+k\d+),\d+)?>"
                ), "characterattributes"))
        tProfile.rules.append((QRegExp(
                r"<\*[lcrj]>"
                r"|<\*t\d+(\.\d+)?(pi)?,[lcrdex](,\"\w\")?>"
                r"|<\*i(\d+(\.\d+)?(\"|in|cm|pt)?(,\d+(\.\d+)?(\"|in|cm|pt)?){0,2})>"
                r"|<\*s\d+(\.\d+)?,\d+(\.\d+)?>"
                r"|<\*S[arf]>"
                r"|<\*t([nbd]\"\w+\",\d+(,\d+)?)?>"
                r"|<\*[hk]\d+>"
                r"|<\*K(n)?>"
                ), "paragraphformats"))
        tProfile.rules.append((QRegExp(
                r"<\\[nitcp1234]>"
                r"|<\\\\>"
                r"|<\\@>"
                r"|<\\<>"
                r"|<\\#\d+>"
                ), "specialcharacters"))
        tProfile.rules.append((QRegExp(
                r"<PageStream\d+\.\d+>"
                ), "metatags"))

        return tProfile
    def pgsResourceProfile(self):
        tProfile = prefsProfile('PgS Resource', "PageStream resource files (*.rsc)", QRegExp(r"\.rsc$"))
        tProfile.fixedPitch = True
        tProfile.tabStop = 4
        tProfile.convertTabs = True
        tProfile.showLineNumbers = False
        tProfile.maintainIndent = 'Normal'
        tProfile.setFormatAttr('normal', '#000000', False, False)
        tProfile.setFormatAttr('stringtable', '#008000', True, False)
        tProfile.setFormatAttr('stringitem', '#20A020', True, False)
        tProfile.setFormatAttr('value', '#800080', True, False)
        tProfile.setFormatAttr('identifier', '#808000', True, False)
        tProfile.setFormatAttr('container', '#800080', True, False)
        tProfile.setFormatAttr('keyword', '#4080A0', True, False)
        tProfile.setFormatAttr('string', '#204080', True, False)
        tProfile.setFormatAttr('object', '#804020', True, False)
        tProfile.setFormatAttr('list', '#408020', True, False)
        tProfile.setFormatAttr('listitem', '#208040', False, True)

        regexp = QRegExp(r"\bSTRINGTABLE\b")
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "stringtable"))
        regexp = QRegExp(r"\bSTRINGITEM\b")
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "stringitem"))
        tProfile.rules.append((QRegExp(r"\b\w\#\d+\b"), "value"))
        tProfile.rules.append((QRegExp(r"\s\#\d+\b"), "identifier"))
        tProfile.rules.append((QRegExp(r"\sL?:\s*\"?[^\"]*\"?"), "string"))
        regexp = QRegExp(
                r"\b[HV]BOX\b"
                r"|\bGRID\b"
                r"|\bBGROUP\b"
                r"|\bREQUESTER\b"
                )
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "container"))
        regexp = QRegExp(
                r"\b(NO)?MARGIN\b"
                r"|\bEND\b"
                r"|\bLABEL\b"
                r"|\bDISPLAY\b"
                r"|\bMINIMENU\b"
                r"|\bSLIDER\b"
                r"|\bLISTVIEW\b"
                r"|\bICON\b"
                r"|\bICONMASK\b"
                r"|\bISOK\b"
                r"|\bISCANCEL\b"
                )
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "keyword"))
        regexp = QRegExp(
                r"\bSTRING\b"
                r"|\bCHECKBOX\b"
                r"|\bBUTTON\b"
                r"|\bPOPUP\b"
                r"|\bHELP\b"
                )
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "object"))
        regexp = QRegExp(
                r"\bLIST\b"
                r"|\bENDLIST\b"
                )
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "list"))
        regexp = QRegExp(
                r"\bLISTITEM\b"
                )
        regexp.setCaseSensitivity(Qt.CaseInsensitive)
        tProfile.rules.append((regexp, "listitem"))
        return tProfile
    def pythonProfile(self):
        KEYWORDS = ["and", "as", "assert", "break", "class", "continue", "def",
                "del", "elif", "else", "except", "exec", "finally", "for", "from",
                "global", "if", "import", "in", "is", "lambda", "not", "or",
                "pass", "print", "raise", "return", "try", "while", "with",
                "yield", "self"]

        BUILTINS = ["abs", "all", "any", "basestring", "bool", "callable", "chr",
                "classmethod", "cmp", "compile", "complex", "delattr", "dict",
                "dir", "divmod", "enumerate", "eval", "execfile", "exit", "file",
                "filter", "float", "frozenset", "getattr", "globals", "hasattr",
                "hex", "id", "int", "isinstance", "issubclass", "iter", "len",
                "list", "locals", "long", "map", "max", "min", "object", "oct",
                "open", "ord", "pow", "property", "range", "reduce", "repr",
                "reversed", "round", "set", "setattr", "slice", "sorted",
                "staticmethod", "str", "sum", "super", "tuple", "type", "unichr",
                "unicode", "vars", "xrange", "zip"] 

        CONSTANTS = ["False", "True", "None", "NotImplemented", "Ellipsis"]

        CLASSKEYS = ["self", "__init__", "__str__", "__call__", "__name__", "__iter__",
                "__get__", "__doc__", "__slots__", "__class__", "__getitem__", "__setitem__",
                "__cmp__", "__contains__", "__delitem__", "__eq__", "__ge__", "__gt__",
                "__lt__", "__le__", "__iter__", "__len__", "__repr__"]

        tProfile = prefsProfile('Python', "Python scripts (*.py;*.pyw)", QRegExp(r"\.pyw?$"))
        tProfile.fixedPitch = True
        tProfile.tabStop = 4
        tProfile.convertTabs = True
        tProfile.maintainIndent = 'Normal'
        tProfile.useBlocks = True
        tProfile.showLineNumbers = True
        tProfile.highlightCurrentLine = True
        tProfile.setFormatAttr("normal", "#000000", False, False)
        tProfile.setFormatAttr("keyword", "#000080", True, False)
        tProfile.setFormatAttr("builtin", "#0000A0", False, False)
        tProfile.setFormatAttr("constant", "#0000C0", False, False)
        tProfile.setFormatAttr("decorator", "#0000E0", False, False)
        tProfile.setFormatAttr("comment", "#007F00", False, True)
        tProfile.setFormatAttr("string", "#808000", False, False)
        tProfile.setFormatAttr("number", "#924900", False, False)
        tProfile.setFormatAttr("error", "#FF0000", False, False)
        tProfile.setFormatAttr("pyqt", "#50621A", False, False)
        tProfile.setFormatAttr("classkeys", "#804040", True, False)
        tProfile.rules.append((QRegExp(
                "|".join([r"\b%s\b" % keyword \
                for keyword in KEYWORDS])),
                "keyword"))
        tProfile.rules.append((QRegExp(
                "|".join([r"\b%s\b" % builtin \
                for builtin in BUILTINS])),
                "builtin"))
        tProfile.rules.append((QRegExp(
                "|".join([r"\b%s\b" % constant \
                for constant in CONSTANTS])),
                "constant"))
        tProfile.rules.append((QRegExp(
                "|".join([r"\b%s\b" % classkey \
                for classkey in CLASSKEYS])),
                "classkeys"))
        tProfile.rules.append((QRegExp(
                r"\b[+-]?[0-9]+[lL]?\b"
                r"|\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b"
                r"|\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b"),
                "number"))
        tProfile.rules.append((QRegExp(
                r"\bPyQt4\b|\bQt?[A-Z][a-z]\w+\b"), "pyqt"))
        tProfile.rules.append((QRegExp(r"\b@\w+\b"), "decorator"))
        tProfile.rules.append((QRegExp(r"^#.*$"), "comment"))
        tProfile.rules.append((QRegExp(r"(?:[^\"'])#.*$"), "comment"))
        stringRe = QRegExp(r"\"[^\"\\\r\n]*(?:\\.[^\"\\\r\n]*)*\"")
        stringRe.setMinimal(True)
        tProfile.rules.append((stringRe, "string"))
        stringRe = QRegExp(r"'[^'\\\r\n]*(?:\\.[^'\\\r\n]*)*'")
        stringRe.setMinimal(True)
        tProfile.rules.append((stringRe, "string"))
        stringRe = QRegExp(r"""(:?"["]".*"["]"|'''.*''')""")
        stringRe.setMinimal(True)
        tProfile.rules.append((stringRe, "string"))
        return tProfile
    def htmlProfile(self):
        tProfile = prefsProfile('HTML', "HTML files (*.html;*.htm)", QRegExp(r"\.html?$"))
        tProfile.fixedPitch = True
        tProfile.tabStop = 4
        tProfile.convertTabs = True
        tProfile.maintainIndent = 'Normal'
        tProfile.showLineNumbers = False
        tProfile.setFormatAttr('normal', '#000000', False, False)
        tProfile.setFormatAttr('entitycodes', '#4080A0', True, False)
        tProfile.setFormatAttr('htmltags', '#808008', True, False)

        tProfile.rules.append((QRegExp(r"&\w+;"), "entitycodes"))
        tProfile.rules.append((QRegExp(r"<[^>]*>"), "htmltags"))
        return tProfile
    def defaultProfile(self):
        tProfile = prefsProfile('default')
        tProfile.setFormatAttr('normal', '#000000', False, False)
        return tProfile

def main():
    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon(":/icon.png"))
    fname = None
    if len(sys.argv) > 1:
        fname = sys.argv[1]
    form = MainWindow(fname)
    form.show()
    all_windows.append(form)
    app.exec_()


if __name__ == "__main__":
    main()
